"""Charliecloud driver as the container runtime system for tasks.

Creates text for tasks using Charliecloud.
"""

import os
from beeflow.common.crt.crt_driver import (ContainerRuntimeDriver, ContainerRuntimeResult)
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.build.build_driver import task2arg
from beeflow.common.container_path import convert_path
from beeflow.common import log as bee_logging


log = bee_logging.setup(__name__)


class CharliecloudDriver(ContainerRuntimeDriver):
    """The ContainerRuntimeDriver for Charliecloud as container runtime system.

    Creates the text for the task for using Charliecloud.
    """

    def __init__(self):
        """Create CharliecloudDriver object."""
        # Retrieve Charlicloud options from configuration file.
        self.chrun_opts = bc.get('charliecloud', 'chrun_opts')
        self.cc_setup = bc.get('charliecloud', 'setup')
        # Read container archive path from config.
        container_archive = bc.get('builder', 'container_archive')
        self.container_archive = bc.resolve_path(container_archive)

    @staticmethod
    def get_ccname(image_path):
        """Strip directories & .tar, .tar.gz, tar.xz, or .tgz from image path."""
        name = os.path.basename(image_path).rsplit('.', 2)
        if name[-1] in ['gz', 'xz']:
            name.pop()
        if name[-1] in ['tar', 'tgz']:
            name.pop()
        name = '.'.join(name)
        return name

    def run_text(self, task):  # noqa
        """Create text for Charliecloud batch script."""
        os.makedirs(self.container_archive, exist_ok=True)
        log.info(f'Build container archive directory is: {self.container_archive}')

        use_container = None
        task_container_name = task.get_requirement('DockerRequirement', 'beeflow:containerName')

        baremetal = False
        if task_container_name is None:
            baremetal = True
            log.info('No beeflow:containerName provided.')
            runtime_target_list = []
            # Harvest beeflow:copyContainer if it exists.
            task_container_path = task.get_requirement('DockerRequirement',
                                                       'beeflow:copyContainer')
            if task_container_path:
                task_container_path = os.path.basename(task_container_path).split('.')[0]
                runtime_target_list.append(task_container_path)

            # Harvest dockerPull if it exists
            task_addr = task.get_requirement('DockerRequirement', 'dockerPull')
            if task_addr:
                task_container_path = task_addr.replace('/', '%')
                runtime_target_list.append(task_container_path)
                log.info(f'Found dockerPull path {task_container_path}. Using its container name.')
                if len(runtime_target_list) > 1:
                    raise RuntimeError(
                        'Too many container runtimes specified! Pick one per workflow step.'
                    )
            if len(runtime_target_list) == 0:
                log.warning('No beeflow:containerName specified.')
            else:
                baremetal = False
                # Build container name from container path.
                task_container_name = runtime_target_list[0]
                log.info(f'Moving w/expectation: {task_container_name} is the container target.')

            # Check for `beeflow:useContainer`
            use_container = task.get_requirement('DockerRequirement', 'beeflow:useContainer')
            if use_container:
                log.info(f'Found beeflow:useContainer option. Using container {use_container}')
                baremetal = False

        # Set the workdir with the env code
        task_workdir_env = f'cd {task.workdir}\n' if task.workdir is not None else ''

        if baremetal:
            return ContainerRuntimeResult(env_code=task_workdir_env, pre_commands=[],
                                          main_command=[str(arg) for arg in task.command],
                                          post_commands=[])

        if task_container_name:
            container_path = '/'.join([self.container_archive, task_container_name]) + '.tar.gz'

        # If use_container is specified, no copying is done, the file  path is used
        if use_container:
            task_container_name = self.get_ccname(use_container)
            container_path = os.path.expanduser(use_container)
        else:
            container_path = '/'.join([self.container_archive, task_container_name]) + '.tar.gz'

        log.info(f'Expecting container at {container_path}. Ready to deploy and run.')

        deployed_image_root = bc.get('builder', 'deployed_image_root')

        hints = dict(task.hints)
        mpi_opt = '--join' if 'beeflow:MPIRequirement' in hints else ''
        command = ' '.join(task.command)
        env_code = ''.join([self.cc_setup if self.cc_setup else '', task_workdir_env])
        deployed_path = deployed_image_root + '/' + task_container_name
        pre_commands = [
            f'mkdir -p {deployed_image_root}\n'.split(),
            f'ch-convert -i tar -o dir {container_path} {deployed_path}\n'.split()
        ]
        # Need to convert the path from inside to outside base on the bind mounts
        extra_opts = ''
        if task.workdir is not None:
            # Only setting it for $HOME right now
            bind_mounts = {
                # Charliecloud bindmounts $HOME to /home/$USER by default
                os.getenv('HOME'): os.path.join('/home', os.getenv('USER')),
            }
            ctr_workdir_path = convert_path(task.workdir, bind_mounts)
            extra_opts = f'--cd {ctr_workdir_path}'
        main_command = (f'ch-run {mpi_opt} {deployed_path} {self.chrun_opts} '
                        f'{extra_opts} -- {command}\n').split()
        post_commands = [
            f'rm -rf {deployed_path}\n'.split(),
        ]
        return ContainerRuntimeResult(env_code, pre_commands, main_command, post_commands)

    def build_text(self, userconfig, task):
        """Build text for Charliecloud batch script."""
        task_args = task2arg(task)
        text = (f'beeflow --build {userconfig} {task_args}\n')
        return text



"""Mid-level interface for managing a build system from WFM.

The WFM may request a Runtime Environment (RTE) that must be built.
This RTE build should be considered a separate stage in the workflow.
The build_interface will access components of the build_driver and
components of the gdb_interface as required.
"""

# from beeflow.common.gdb.gdb_interface import GraphDatabaseInterface
# from beeflow.common.build.container_drivers import CharliecloudBuildDriver,
#                                                    SingularityBuildDriver
import sys
from subprocess import CalledProcessError
from beeflow.common.build.container_drivers import CharliecloudBuildDriver
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common import log as bee_logging
from beeflow.common.build.build_driver import arg2task


log = bee_logging.setup(__name__)


def build_main(task):
    """Process build instructions - main code."""
    # The build driver treats Hint and Requirement objects as Dicts.
    task_local = task.copy()
    task_local.hints = dict(task_local.hints)
    task_local.requirements = dict(task_local.requirements)
    builder = CharliecloudBuildDriver(task_local)
    log.info('CharliecloudBuildDriver initialized')

    # Deque the next build instruction until empty or reach terminal case
    # ...catch the case where no build instructions are required.
    op_keys = ["build_op", "op_name", "op_priority", "op_terminal"]
    try:
        op_values = builder.exec_list.pop(0)
    except IndexError:
        log.info('No build instructions provided. Assuming this is ok...')
        op_values = [None, None, None, True]
    op_dict = dict(zip(op_keys, op_values))
    while op_dict["build_op"]:
        log.info(f'Executing build operation: {op_dict["op_name"]}')
        try:
            return_obj = op_dict["build_op"]()
            # Return objects will be successful subprocess or return code.
            try:
                return_code = return_obj.returncode
            except AttributeError:
                return_code = int(return_obj)
        except CalledProcessError:
            return_code = 1
            log.warning(f'There was a problem executing {op_dict["op_name"]}!')
        # Case 1: Not the last operation spec'd, but is a terminal operation.
        if op_dict["op_terminal"] and return_code == 0:
            op_values = [None, None, None, True]
            op_dict = dict(zip(op_keys, op_values))
            continue
        # Case 2: Go to next, or last operation spec'd.
        try:
            op_values = builder.exec_list.pop(0)
        except IndexError:
            op_values = [None, None, None, True]
        finally:
            op_dict = dict(zip(op_keys, op_values))


if __name__ == '__main__':
    try:
        bc.init(userconfig=sys.argv[1])
        my_args = sys.argv[2]
    except IndexError as exc:
        raise IndexError('build_interface must execute with 2 arguments.') from exc

    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    handler = bee_logging.save_log(bee_workdir=bee_workdir, log=log, logfile='build_interface.log')

    try:
        local_task = arg2task(my_args)
    except Exception as err:
        log.info(f'{err}')

    build_main(local_task)

# Ignore W0703: Catching generic exception isn't a problem if we just want a descriptive report
# pylama:ignore=W0703



"""OpenStack provider module."""
import openstack
from beeflow.common.cloud import provider
from beeflow.common.cloud.cloud import CloudError


class OpenstackProvider(provider.Provider):
    """OpenStack provider class."""

    def __init__(self, stack_name, **_kwargs):
        """Construct a new OpenStack provider class."""
        self._cloud = openstack.connect()
        self._stack_name = stack_name

    def get_ext_ip_addr(self, node_name):
        """Get external IP address of Task Manager node."""
        node = self._cloud.get_server(node_name)
        if node is None:
            raise CloudError('Cannot retrieve node/IP information. Is `node_name` set correctly?')
        return node.accessIPv4

    def setup_cloud(self, config):
        """Set up the cloud based on config data."""
        # Just write out the template to the pwd for right now
        template_file = './openstack.yaml'
        with open(template_file, 'w', encoding='utf-8') as fp:
            fp.write(config)
        self._cloud.create_stack(self._stack_name, template_file=template_file, wait=True)



"""Abstract base class for worker, the workload manager."""

from abc import ABC, abstractmethod
import os
import jinja2
from beeflow.common import log as bee_logging
from beeflow.common.crt_interface import ContainerRuntimeInterface


log = bee_logging.setup(__name__)


# Import all implemented container runtime drivers now
# No error if they don't exist
try:
    from beeflow.common.crt.charliecloud_driver import CharliecloudDriver
except ModuleNotFoundError:
    pass
try:
    from beeflow.common.crt.singularity_driver import SingularityDriver
except ModuleNotFoundError:
    pass


class WorkerError(Exception):
    """Worker error class."""

    def __init__(self, *args):
        """Construct a new worker error."""
        super(Exception, self).__init__()
        self.args = args


class Worker(ABC):
    """Worker interface for a generic workload manager."""

    def __init__(self, bee_workdir, **kwargs):
        """Load appropriate container runtime driver, based on configs in kwargs."""
        try:
            self.tm_crt = kwargs['container_runtime']
            if self.tm_crt == 'Charliecloud':
                crt_driver = CharliecloudDriver
            elif self.tm_crt == 'Singularity':
                crt_driver = SingularityDriver
            self.crt = ContainerRuntimeInterface(crt_driver)
        except KeyError:
            log.warning("No container runtime specified in config, proceeding with caution.")
            self.tm_crt = None
            crt_driver = None

        # Get BEE workdir from config file
        self.workdir = bee_workdir

        # Get template for job
        self.job_template = kwargs['job_template']
        if self.job_template:
            # Make sure that the file exists and is readable
            try:
                with open(self.job_template, 'r', encoding='UTF-8') as template_file:
                    template_file.read()
                log.info(f'Jobs will use template: {self.job_template}')
            except ValueError as error:
                raise RuntimeError(f'Cannot open {self.job_template}, {error}') from error
            except FileNotFoundError as exc:
                raise RuntimeError(f'Cannot find job template {self.job_template}') from exc
            except PermissionError as error:
                raise RuntimeError(f'Permission error {self.job_template}') from error
        else:
            raise RuntimeError('No job_template found for the worker class')

    @property
    def template(self):
        """Load the template file."""
        with open(self.job_template, encoding='UTF-8') as fp:
            return jinja2.Template(fp.read())

    def task_save_path(self, task):
        """Return the task save path used for storing submission scripts output logs."""
        return f'{self.workdir}/workflows/{task.workflow_id}/{task.name}-{task.id}'

    def prepare(self, task):
        """Prepare for the task; create the task save directory, etc."""
        task_save_path = self.task_save_path(task)
        os.makedirs(task_save_path, exist_ok=True)

    @abstractmethod
    def build_text(self, task):
        """Build text for task script; use template if it exists.

        :param task: task that we're building a script for
        :type task: Task
        :rtype: string
        """

    @abstractmethod
    def submit_task(self, task):
        """Worker submits task; returns job_id, job_state.

        :param task: instance of Task
        :rtype: tuple (int, string)
        """

    @abstractmethod
    def cancel_task(self, job_id):
        """Cancel task with job_id; returns job_state.

        :param job_id: to be cancelled
        :type job_id: integer
        :rtype: string
        """

    @abstractmethod
    def query_task(self, job_id):
        """Query job state for the task.

        :param job_id: job id to query for status.
        :type job_id: int
        :rtype: string
        """

# Ignore W0511: Ignore TODOs since we will want to address this later.
# pylama:ignore=W0511



"""Start up the workflow manager connecting all of the endpoints."""

from flask import Flask
from beeflow.common.api import BeeApi

from beeflow.wf_manager.resources.wf_list import WFList
from beeflow.wf_manager.resources.wf_actions import WFActions
from beeflow.wf_manager.resources.wf_update import WFUpdate

from beeflow.wf_manager.resources import wf_utils


def create_app():
    """Create flask app object and add REST endpoints."""
    app = Flask(__name__)
    api = BeeApi(app)

    # Add endpoints
    api.add_resource(WFList, '/bee_wfm/v1/jobs/')
    api.add_resource(WFActions, '/bee_wfm/v1/jobs/<string:wf_id>')
    api.add_resource(WFUpdate, '/bee_wfm/v1/jobs/update/')
    return app


if __name__ == '__main__':
    flask_app = create_app()
    bee_workdir = wf_utils.get_bee_workdir()
    # handler = bee_logging.save_log(bee_workdir=bee_workdir, log=log, logfile='wf_manager.log')
    # wfm_listen_port = wf_utils.get_open_port()
    # wf_db.set_wfm_port(wfm_listen_port)
    # flask_app.run(debug=False, port=str(wfm_listen_port))



"""CWL workflow description parser.

Parses the contents of a CWL file and generates the graph representation of parsed workflows and
tools in the graph database. This parser supports a subset of the CWL v1.0 standard. It is
inspired by the CWL parser written by the CWL parser written by the SABER team at John Hopkins
University (see SABER project at https://github.com/aplbrain/saber).
"""
import sys
import argparse
import json
import os
import traceback
import yaml
import cwl_utils.parser.cwl_v1_2 as cwl_parser
from schema_salad.exceptions import ValidationException  # noqa (pylama can't find the exception)

from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.wf_data import (InputParameter,
                                    OutputParameter,
                                    StepInput,
                                    StepOutput,
                                    Hint,
                                    Requirement,
                                    generate_workflow_id)
from beeflow.wf_manager.resources import wf_utils


if not bc.ready():
    bc.init()


# Map CWL types to Python types
type_map = {
    "string": str,
    "int": int,
    "long": int,
    "float": float,
    "double": float,
    "File": str,
    "Directory": str,
}


class CwlParseError(Exception):
    """Parser error class."""

    def __init__(self, *args):
        """Create a parser error."""
        self.args = args


class CwlParser:
    """Class for parsing CWL files."""

    def __init__(self, wfi):
        """Initialize the CWL parser interface.

        Sets the workflow interface for communication with the graph database.
        """
        self.cwl = None
        self.path = None
        self.steps = []
        self.params = None
        self._wfi = wfi

    def parse_workflow(self, workflow_id, cwl_path, job=None):
        """Parse a CWL Workflow file and load it into the graph database.

        Returns an instance of the WorkflowInterface.

        :param workflow_id: the workflow ID
        :type workflow_id: str
        :param cwl_path: the CWL file path
        :type cwl_path: str
        :param job: the input job file (YAML or JSON)
        :type job: str
        :rtype: WorkflowInterface
        """
        self.path = cwl_path
        try:
            self.cwl = cwl_parser.load_document(cwl_path)
        except ValidationException as err:
            traceback.print_exc()
            raise CwlParseError(*err.args) from None

        if self.cwl.class_ != "Workflow":
            raise CwlParseError(f"{os.path.basename(cwl_path)} class must be Workflow")

        if job:
            # Parse input job params into self.params
            self.parse_job(job)
        else:
            self.params = {}

        def resolve_input(input_, type_):
            """Resolve workflow input parameter from job file.

            :param input_: the workflow input name
            :type input_: WorkflowInputParameter
            :param type_: the workflow input type
            :type type_: str
            :rtype: str or int or float
            """
            # Use parsed input parameter for input value if it exists
            input_id = _shortname(input_.id)
            if input_id not in self.params:
                return None
            if not isinstance(self.params[input_id], type_map[type_]):
                raise CwlParseError("Input/param types do not match: "
                                    f"{input_id}/{self.params[input_id]}")
            return self.params[input_id]

        workflow_name = os.path.basename(cwl_path).split(".")[0]
        workflow_inputs = {InputParameter(_shortname(input_.id), input_.type,
                                          resolve_input(input_, input_.type))
                           for input_ in self.cwl.inputs}
        workflow_outputs = {OutputParameter(_shortname(output.id), output.type, None,
                                            _shortname(output.outputSource, True))
                            for output in self.cwl.outputs}
        workflow_hints = self.parse_requirements(self.cwl.hints, as_hints=True)
        workflow_requirements = self.parse_requirements(self.cwl.requirements)

        self._wfi.initialize_workflow(workflow_id, workflow_name, workflow_inputs,
                                      workflow_outputs, workflow_requirements, workflow_hints)
        for step in self.cwl.steps:
            self.parse_step(step)

        return self._wfi

    def parse_step(self, step):
        """Parse a CWL step object.

        Calling this to parse a CommandLineTool file without a corresponding
        Workflow file will fail.

        :param step: the CWL step object
        :type step: WorkflowStep
        """
        # Parse CWL file specified by run field, else parse run field as inline CommandLineTool
        if isinstance(step.run, str):
            step_run = f"{os.path.dirname(step.id)}/{step.run}"
            step_cwl = cwl_parser.load_document(step_run)
            step_id = os.path.basename(step_cwl.id).split(".")[0]
        else:
            step_cwl = step.run
            step_id = _shortname(step.id)

        if step_cwl.class_ != "CommandLineTool":
            raise CwlParseError(f"Step {step.id} class must be CommandLineTool")

        step_name = os.path.basename(step_id).split(".")[0]
        step_command = step_cwl.baseCommand
        step_inputs = self.parse_step_inputs(step.in_, step_cwl.inputs)
        step_outputs = self.parse_step_outputs(step.out, step_cwl.outputs, step_cwl.stdout,
                                               step_cwl.stderr)
        step_requirements = self.parse_requirements(step.requirements)
        step_requirements.extend(self.parse_requirements(step_cwl.requirements))
        step_hints = self.parse_requirements(step.hints, as_hints=True)
        step_hints.extend(self.parse_requirements(step_cwl.hints, as_hints=True))
        step_stdout = step_cwl.stdout
        step_stderr = step_cwl.stderr

        self._wfi.add_task(step_name, base_command=step_command, inputs=step_inputs,
                           outputs=step_outputs, requirements=step_requirements,
                           hints=step_hints, stdout=step_stdout, stderr=step_stderr)

    def parse_job(self, job):
        """Parse a CWL input job file.

        Input parameters are stored in the params attribute.

        :param job: the path of the input job file (YAML or JSON)
        :type job: str
        """
        if job.endswith(".yml") or job.endswith(".yaml"):
            with open(job, encoding="utf-8") as fp:
                self.params = yaml.full_load(fp)
        elif job.endswith(".json"):
            with open(job, encoding="utf-8") as fp:
                self.params = json.load(fp)
        else:
            raise CwlParseError("Unsupported input job file extension (only .yml "
                                "and .json supported)")

        for k, v in self.params.items():
            if not isinstance(k, str):
                raise CwlParseError(f"Invalid input job key: {str(k)}")
            if not isinstance(v, (str, int, float)):
                raise CwlParseError(f"Invalid input job parameter type: {type(v)}")

    @staticmethod
    def parse_step_inputs(cwl_in, step_inputs):
        """Parse step inputs from CWL step input objects.

        :param cwl_in: the step inputs from the Workflow file
        :type cwl_in: list of WorkflowStepInput
        :param step_inputs: the step inputs from the CommandLineTool file
        :type step_inputs: list of CommandInputParameter
        :rtype: list of StepInput
        """
        source_map = {_shortname(input_.id).split("/")[-1]: _shortname(input_.source)
                      for input_ in cwl_in}

        inputs = []
        for step_input in step_inputs:
            # If the input type is str, then the input is required
            # If it is a list containing 'null' and another type(s) then it is optional
            # If the input is not in the source_map but has a default value then it will
            # be set by the GDB later
            if _shortname(step_input.id) not in source_map.keys():
                if isinstance(step_input.type, str) and step_input.default is None:
                    raise CwlParseError(f"required input {_shortname(step_input.id)} "
                                        "not satisfied")
                continue

            if isinstance(step_input.type, str):
                input_type = step_input.type
            else:
                input_type = step_input.type[1]

            if step_input.inputBinding:
                inputs.append(StepInput(_shortname(step_input.id), input_type, None,
                                        step_input.default, source_map[_shortname(step_input.id)],
                                        step_input.inputBinding.prefix,
                                        step_input.inputBinding.position,
                                        step_input.inputBinding.valueFrom))
            else:
                inputs.append(StepInput(_shortname(step_input.id), input_type, None,
                                        step_input.default, source_map[_shortname(step_input.id)],
                                        None, None))

        return inputs

    @staticmethod
    def parse_step_outputs(cwl_out, step_outputs, stdout, stderr):
        """Parse step outputs from CWL step output objects.

        :param cwl_out: the step outputs from the Workflow file
        :type cwl_out: list of str
        :param step_outputs: the step outputs from the CommandLineTool file
        :type step_outputs: list of CommandOutputParameter
        :param stdout: name of file to which stdout should be redirected
        :type stdout: str or None
        :param stderr: name of file to which stderr should be redirected
        :type stderr: str or None
        :rtype: list of StepOutput
        """
        out_short = list(map(_shortname, cwl_out))
        short_id = out_short[0].split("/")[0]
        # Inline step outputs already have short_id+"/" prepended
        out_map = {(_shortname(step_output.id)
                   if _shortname(step_output.id).startswith(short_id + "/") else
                   short_id + "/" + _shortname(step_output.id)): step_output
                   for step_output in step_outputs}

        outputs = []
        for out in out_short:
            if out not in out_map.keys():
                raise CwlParseError(f"specified step output {out} not produced by CommandLineTool")

            output_type = out_map[out].type
            glob = None
            if output_type == "stdout":
                if not stdout:
                    raise CwlParseError(f"stdout capture required for step output {out} "
                                        "but not specified by CommandLineTool")
                # Fill in glob with value of stdout
                glob = stdout
            elif output_type == "stderr":
                if not stderr:
                    raise ValueError((f"stderr capture required for step output {out} "
                                      "but not specified by CommandLineTool"))
                # Fill in glob with value of stderr
                glob = stderr
            else:
                if out_map[out].outputBinding:
                    glob = out_map[out].outputBinding.glob

            outputs.append(StepOutput(out, output_type, None, glob))

        return outputs

    def parse_requirements(self, requirements, as_hints=False):
        """Parse CWL hints/requirements.

        :param requirements: the CWL requirements
        :type requirements: list of ordereddict or any cwl_utils Requirement class
        :param as_hints: parse as hints instead of requirements
        :type as_hints: bool
        :rtype: list of Hint or list of Requirement or None
        """
        reqs = []
        if not requirements:
            return reqs
        if as_hints:
            for req in requirements:
                items = {k: v for k, v in req.items() if k != "class"}
                # Load in the dockerfile at parse time
                if 'dockerFile' in items:
                    base_path = os.path.dirname(self.path)
                    docker_file = items['dockerFile']
                    path = os.path.join(base_path, docker_file)
                    try:
                        with open(path, encoding='utf-8') as fp:
                            items['dockerFile'] = fp.read()
                    except FileNotFoundError:
                        msg = f'Could not find the docker file: {docker_file}'
                        raise CwlParseError(msg) from None
                reqs.append(Hint(req['class'], items))
        else:
            for req in requirements:
                reqs.append(Requirement(req.class_, {k: v for k, v in vars(req).items()
                                                     if k not in ("extension_fields",
                                                                  "loadingOptions", "class_")
                                                     and v is not None}))
        return reqs


def _shortname(uri, output_source=False):
    """Shorten a CWL object URI.

    e.g., file:///path/to/file#step/name -> step/name,
    or file:///path/to/file#output/output/source -> output/source if outputSource is True

    :param uri: a CWL object URI
    :type uri: str
    :param output_source: true if URI is for an outputSource object, else false
    :type output_source: bool
    """
    if output_source:
        output = uri.split("#")[-1]
        return "/".join(output.split("/")[1:])

    return uri.split("#")[-1]


def parse_args(args=None):
    """Parse arguments."""
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)

    parser.add_argument("wf_file", type=str, help="CWL workflow file")
    parser.add_argument("-i", "--inputs", type=str, help="Workflow job inputs file",
                        required=False)

    return parser.parse_args(args)


def main():
    """Run the parser on a CWL Workflow and job file directly."""
    wf_id = generate_workflow_id()
    wfi = wf_utils.get_workflow_interface(wf_id)
    parser = CwlParser(wfi)
    args = parse_args()
    parser.parse_workflow(wf_id, args.wf_file, args.inputs)


if __name__ == "__main__":
    sys.exit(main())



"""Contains the workflow update REST endpoint."""

import os
import json
import shutil
import subprocess
import time
import jsonpickle

from flask import make_response, jsonify
from flask_restful import Resource, reqparse
from beeflow.wf_manager.common import wf_db
from beeflow.wf_manager.resources import wf_utils
from beeflow.wf_manager.common import dep_manager
from beeflow.common import log as bee_logging

log = bee_logging.setup(__name__)


def archive_workflow(wf_id):
    """Archive a workflow after completion."""
    # Archive Config
    workflow_dir = wf_utils.get_workflow_dir(wf_id)
    shutil.copyfile(os.path.expanduser("~") + '/.config/beeflow/bee.conf',
                    workflow_dir + '/' + 'bee.conf')

    wf_db.update_workflow_state(wf_id, 'Archived')
    wf_utils.update_wf_status(wf_id, 'Archived')

    bee_workdir = wf_utils.get_bee_workdir()
    archive_dir = os.path.join(bee_workdir, 'archives')
    os.makedirs(archive_dir, exist_ok=True)
    archive_path = f'../archives/{wf_id}.tgz'
    # We use tar directly since tarfile is apparently very slow
    workflows_dir = wf_utils.get_workflows_dir()
    subprocess.call(['tar', '-czf', archive_path, wf_id], cwd=workflows_dir)


class WFUpdate(Resource):
    """Class to interact with an existing workflow."""

    def __init__(self):
        """Set up arguments."""
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument('wf_id', type=str, location='json',
                                   required=True)
        self.reqparse.add_argument('task_id', type=str, location='json',
                                   required=True)
        self.reqparse.add_argument('job_state', type=str, location='json',
                                   required=True)
        self.reqparse.add_argument('metadata', type=str, location='json',
                                   required=False)
        self.reqparse.add_argument('task_info', type=str, location='json',
                                   required=False)
        self.reqparse.add_argument('output', location='json', required=False)

    def put(self):
        """Update the state of a task from the task manager."""
        data = self.reqparse.parse_args()
        wf_id = data['wf_id']
        task_id = data['task_id']
        job_state = data['job_state']

        wfi = wf_utils.get_workflow_interface(wf_id)
        task = wfi.get_task_by_id(task_id)
        wfi.set_task_state(task, job_state)
        wf_db.update_task_state(task_id, wf_id, job_state)
        # wf_profiler.add_state_change(task, job_state)

        # Get metadata from update if available
        if 'metadata' in data:
            if data['metadata'] is not None:
                metadata = jsonpickle.decode(data['metadata'])
                wfi.set_task_metadata(task, metadata)

        bee_workdir = wf_utils.get_bee_workdir()
        # Get output from the task
        if 'metadata' in data:
            if data['metadata'] is not None:
                metadata = jsonpickle.decode(data['metadata'])
                old_metadata = wfi.get_task_metadata(task)
                old_metadata.update(metadata)
                wfi.set_task_metadata(task, old_metadata)

        if 'output' in data and data['output'] is not None:
            fname = f'{wfi.workflow_id}_{task.id}_{int(time.time())}.json'
            task_output_path = os.path.join(bee_workdir, fname)
            with open(task_output_path, 'w', encoding='utf8') as fp:
                json.dump(json.loads(data['output']), fp, indent=4)

        if 'task_info' in data and data['task_info'] is not None:
            task_info = jsonpickle.decode(data['task_info'])
            checkpoint_file = task_info['checkpoint_file']
            new_task = wfi.restart_task(task, checkpoint_file)
            if new_task is None:
                log.info('No more restarts')
                state = wfi.get_task_state(task)
                return make_response(jsonify(status=f'Task {task_id} set to {job_state}'))
            # Submit the restart task
            tasks = [new_task]
            wf_utils.schedule_submit_tasks(wf_id, tasks)
            return make_response(jsonify(status='Task {task_id} restarted'))

        if job_state in ('COMPLETED', 'FAILED'):
            for output in task.outputs:
                if output.glob is not None:
                    wfi.set_task_output(task, output.id, output.glob)
                else:
                    wfi.set_task_output(task, output.id, "temp")
            tasks = wfi.finalize_task(task)
            state = wfi.get_workflow_state()
            if tasks and state != 'PAUSED':
                wf_utils.schedule_submit_tasks(wf_id, tasks)

            if wfi.workflow_completed():
                log.info("Workflow Completed")
                # Save the profile
                # wf_profiler.save()
                wf_id = wfi.workflow_id
                archive_workflow(wf_id)
                pid = wf_db.get_gdb_pid(wf_id)
                dep_manager.kill_gdb(pid)

        resp = make_response(jsonify(status=(f'Task {task_id} belonging to WF {wf_id} set to'
                                             f'{job_state}')), 200)
        return resp



"""Mid-level interface for managing a graph database with workflows.

Delegates the actual work to an instance of a subclass of
the abstract base class GraphDatabaseDriver. By default,
this is the Neo4jDriver class.
"""

from beeflow.common.gdb.neo4j_driver import Neo4jDriver
from beeflow.common import expr


class GraphDatabaseInterface:
    """Interface for managing a graph database with workflows.

    Requires an implemented subclass of GraphDatabaseDriver (uses Neo4jDriver by default).
    """

    def __init__(self, gdb_driver=Neo4jDriver):
        """Initialize the graph database interface with a graph database driver.

        :param gdb_driver: the graph database driver (Neo4jDriver by default)
        :type gdb_driver: subclass of GraphDatabaseDriver
        """
        # Store the GDB driver state
        self._connection = gdb_driver

    def __del__(self):
        """Disconnect from the graph database when interface deconstructs."""
        if self.initialized():
            self.close()

    def connect(self, **kwargs):
        """Initialize a graph database interface with a driver.

        :param kwargs: arguments for initializing the graph database connection
        """

    def initialize_workflow(self, workflow):
        """Begin construction of a workflow in the graph database.

        Connects to the database and creates the Workflow with accompanying Hint
        and Requirement nodes.
        Permits the addition of task nodes to the workflow.

        :param workflow: the workflow description
        :type workflow: Workflow
        """
        self._connection.initialize_workflow(workflow)

    def execute_workflow(self):
        """Begin execution of the loaded workflow."""
        self._connection.execute_workflow()

    def pause_workflow(self):
        """Pause execution of a running workflow."""
        self._connection.pause_workflow()

    def resume_workflow(self):
        """Resume execution of a running workflow."""
        self._connection.resume_workflow()

    def reset_workflow(self, new_id):
        """Reset the execution state and ID of a workflow.

        :param new_id: the new workflow ID
        :type new_id: str
        """
        self._connection.reset_workflow(new_id)

    def load_task(self, task):
        """Load a task into a workflow in the graph database.

        :param task: the task
        :type task: Task
        """
        self._connection.load_task(task)

    def initialize_ready_tasks(self):
        """Set runnable tasks in a workflow to ready."""
        self._connection.initialize_ready_tasks()

    def restart_task(self, old_task, new_task):
        """Create a new task from a failed task checkpoint restart enabled.

        :param old_task: the failed task
        :type old_task: Task
        :param new_task: the new (restarted) task
        :type new_task: Task
        """
        self._connection.restart_task(old_task, new_task)

    def finalize_task(self, task):
        """Set a task's state to completed.

        :param task: the task to finalize
        :type task: Task
        """
        self._connection.finalize_task(task)

    def get_task_by_id(self, task_id):
        """Return a workflow Task given its ID.

        :param task_id: the task's ID
        :type task_id: str
        :rtype: Task
        """
        return self._connection.get_task_by_id(task_id)

    def get_workflow_description(self):
        """Return the workflow description from the graph database.

        :rtype: Workflow
        """
        return self._connection.get_workflow_description()

    def get_workflow_state(self):
        """Return workflow's current state.

        :rtype: str
        """
        return self._connection.get_workflow_state()

    def set_workflow_state(self, state):
        """Return workflow's current state.

        :param state: the new state of the workflow
        :type state: str
        """
        self._connection.set_workflow_state(state)

    def get_workflow_tasks(self):
        """Return a workflow's tasks from the graph database.

        :rtype: list of Task
        """
        return self._connection.get_workflow_tasks()

    def get_workflow_requirements_and_hints(self):
        """Return a tuple containing a list of requirements and a list of hints.

        The returned tuple format is (requirements, hints).

        :rtype: (list of Requirement, list of Hint)
        """
        return self._connection.get_workflow_requirements_and_hints()

    def get_ready_tasks(self):
        """Return the tasks in a workflow with state 'READY'.

        :rtype: list of Task
        """
        return self._connection.get_ready_tasks()

    def get_dependent_tasks(self, task):
        """Return the dependents of a task in a workflow.

        :param task: the task whose dependents to retrieve
        :type task: Task
        :rtype: list of Task
        """
        return self._connection.get_dependent_tasks(task)

    def get_task_state(self, task):
        """Return the state of a task.

        :param task: the task whose state to retrieve
        :type task: Task
        :rtype: str
        """
        return self._connection.get_task_state(task)

    def set_task_state(self, task, state):
        """Set the state of a task.

        :param task: the task whose state to set
        :type task: Task
        :param state: the new state
        :type state: str
        """
        return self._connection.set_task_state(task, state)

    def get_task_metadata(self, task):
        """Return the job description metadata of a task.

        :param task: the task whose metadata to retrieve
        :type task: Task
        :rtype: dict
        """
        return self._connection.get_task_metadata(task)

    def set_task_metadata(self, task, metadata):
        """Set the job description metadata of a task.

        :param task: the task whose metadata to set
        :type task: Task
        :param metadata: the job description metadata
        :type metadata: dict
        """
        self._connection.set_task_metadata(task, metadata)

    def get_task_input(self, task, input_id):
        """Get a task input object.

        :param task: the task whose input to retrieve
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :rtype: StepInput
        """
        return self._connection.get_task_input(task, input_id)

    def set_task_input(self, task, input_id, value):
        """Set the value of a task input.

        :param task: the task whose input to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param value: str or int or float
        """
        self._connection.set_task_input(task, input_id, value)

    def get_task_output(self, task, output_id):
        """Get a task output object.

        :param task: the task whose output to retrieve
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :rtype: StepOutput
        """
        return self._connection.get_task_output(task, output_id)

    def set_task_output(self, task, output_id, value):
        """Set the value of a task output.

        :param task: the task whose output to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param value: the output value to set
        :type value: str or int or float
        """
        self._connection.set_task_output(task, output_id, value)

    def evaluate_expression(self, task, id_, output):
        """Evaluate a task input/output expression.

        Expression can be a parameter substitution or a string concatenation in a StepInput
        valueFrom field or a StepOutput glob field. The only special variable supported
        is self.path.

        :param task: the task whose expression to evaluate
        :type task: Task
        :param id_: the id of the step input/output
        :type id_: str
        :param output: true if output glob expression being evaluated, else false
        :type output: bool
        """
        # Get ground-truth task from database before doing anything
        task = self._connection.get_task_by_id(task.id)
        input_pairs = {input.id: input.value for input in task.inputs}
        if output:
            step_output = self._connection.get_task_output(task, id_)
            val = expr.eval_output(input_pairs, step_output.glob)
            if val is not None:
                self._connection.set_task_output_glob(task, id_, val)
        else:
            step_input = self._connection.get_task_input(task, id_)
            val = expr.eval_input(input_pairs, step_input.value_from)
            self._connection.set_task_input_type(task, id_, "string")
            self._connection.set_task_input(task, id_, val)

    def workflow_completed(self):
        """Return true if all of a workflow's final tasks have completed, else false.

        :rtype: bool
        """
        return self._connection.workflow_completed()

    def initialized(self):
        """Return true if the database connection has been initialized, else false.

        :rtype: bool
        """
        return bool(self._connection is not None)

    def empty(self):
        """Return true if the graph database is empty, else false.

        :rtype: bool
        """
        return self._connection.empty()

    def cleanup(self):
        """Clean up all data in the graph database."""
        self._connection.cleanup()

    def close(self):
        """Close the connection to the graph database."""
        self._connection.close()
        self._connection = None



"""BEE Scheduler tests.

Tests of the BEE Scheduler module.
"""
from beeflow.scheduler import algorithms
from beeflow.scheduler import task
from beeflow.scheduler import resource_allocation


class TestFCFS:
    """Test the FCFS algorithm.

    Test the FCFS algorithm.
    """

    @staticmethod
    def test_schedule_one_task():
        """Test scheduling one task."""
        schedule_one_task(algorithms.FCFS())

    @staticmethod
    def test_schedule_two_tasks():
        """Test scheduling two tasks."""
        schedule_two_tasks(algorithms.FCFS())

    @staticmethod
    def test_schedule_task_fail():
        """Test scheduling a task with more resources required than available."""
        schedule_task_fail(algorithms.FCFS())

    @staticmethod
    def test_schedule_task_gpus_req():
        """Test scheduling a task with a gpus requirement."""
        schedule_task_gpus_req(algorithms.FCFS())

    @staticmethod
    def test_schedule_task_gpus_req_fail():
        """Test scheduling a task with a gpus requirement that should fail."""
        schedule_task_gpus_req_fail(algorithms.FCFS())

    @staticmethod
    def test_schedule_six_tasks():
        """Test scheduling six tasks."""
        requirements1 = {'max_runtime': 3}
        task1 = task.Task(workflow_name='workflow-1', task_name='task-1',
                          requirements=requirements1)
        task2 = task.Task(workflow_name='workflow-1', task_name='task-2',
                          requirements=requirements1)
        requirements2 = {'max_runtime': 4}
        task3 = task.Task(workflow_name='workflow-1', task_name='task-3',
                          requirements=requirements2)
        task4 = task.Task(workflow_name='workflow-1', task_name='task-4',
                          requirements=requirements2)
        task5 = task.Task(workflow_name='workflow-1', task_name='task-5',
                          requirements=requirements2)
        task6 = task.Task(workflow_name='workflow-1', task_name='task-6',
                          requirements=requirements2)
        resource = resource_allocation.Resource(id_='test-resource-1', nodes=4)

        tasks = [task1, task2, task3, task4, task5, task6]
        algorithms.FCFS().schedule_all(tasks, [resource])

        assert task1.allocations[0].id_ == 'test-resource-1'
        assert task1.allocations[0].start_time == 0
        assert task1.allocations[0].nodes == 1
        assert task2.allocations[0].id_ == 'test-resource-1'
        assert task2.allocations[0].start_time == 0
        assert task2.allocations[0].nodes == 1
        assert task3.allocations[0].id_ == 'test-resource-1'
        assert task3.allocations[0].start_time == 0
        assert task3.allocations[0].nodes == 1
        assert task4.allocations[0].id_ == 'test-resource-1'
        assert task4.allocations[0].start_time == 0
        assert task4.allocations[0].nodes == 1
        assert task5.allocations[0].id_ == 'test-resource-1'
        t = task2.requirements.max_runtime
        assert task5.allocations[0].start_time == t
        assert task5.allocations[0].nodes == 1
        assert task6.allocations[0].id_ == 'test-resource-1'
        assert task6.allocations[0].start_time == t
        assert task6.allocations[0].nodes == 1


class TestBackfill:
    """Test the Backfill algorithm."""

    @staticmethod
    def test_schedule_one_task():
        """Test scheduling one task."""
        schedule_one_task(algorithms.Backfill())

    @staticmethod
    def test_schedule_two_tasks():
        """Test scheduling two tasks."""
        schedule_two_tasks(algorithms.Backfill())

    @staticmethod
    def test_schedule_task_fail():
        """Test scheduling a task with more resources required than available."""
        schedule_task_fail(algorithms.Backfill())

    @staticmethod
    def test_schedule_task_gpus_req():
        """Test scheduling a task with a gpus requirement."""
        schedule_task_gpus_req(algorithms.Backfill())

    @staticmethod
    def test_schedule_task_gpus_req_fail():
        """Test scheduling a task with a gpus requirement that should fail."""
        schedule_task_gpus_req_fail(algorithms.Backfill())

    @staticmethod
    def test_schedule_three_tasks():
        """Test scheduling three tasks."""
        requirements = {'max_runtime': 1, 'nodes': 1}
        task1 = task.Task(workflow_name='workflow-0', task_name='task-1',
                          requirements=requirements)
        requirements = {'max_runtime': 1, 'nodes': 2}
        task2 = task.Task(workflow_name='workflow-0', task_name='task-2',
                          requirements=requirements)
        requirements = {'max_runtime': 1, 'nodes': 1}
        task3 = task.Task(workflow_name='workflow-0', task_name='task-3',
                          requirements=requirements)
        resource = resource_allocation.Resource(id_='resource-0', nodes=2)

        tasks = [task1, task2, task3]
        algorithms.Backfill().schedule_all(tasks, [resource])

        assert task1.allocations[0].id_ == 'resource-0'
        assert task1.allocations[0].nodes == 1
        assert task1.allocations[0].start_time == 0
        assert task2.allocations[0].id_ == 'resource-0'
        assert task2.allocations[0].nodes == 2
        assert (task2.allocations[0].start_time
                == task1.requirements.max_runtime)
        # Task 3 should have been backfillled, filling in an area before task
        # 2 can run
        assert task3.allocations[0].id_ == 'resource-0'
        assert task3.allocations[0].nodes == 1
        assert task3.allocations[0].start_time == 0

    @staticmethod
    def test_schedule_four_tasks():
        """Test scheduling four tasks."""
        requirements = {'max_runtime': 2, 'nodes': 4}
        task1 = task.Task(workflow_name='workflow-0', task_name='task-0',
                          requirements=requirements)
        requirements = {'max_runtime': 2, 'nodes': 8}
        task2 = task.Task(workflow_name='workflow-0', task_name='task-1',
                          requirements=requirements)
        requirements = {'max_runtime': 3, 'nodes': 2}
        # This task should not be backfilled (too much time)
        task3 = task.Task(workflow_name='workflow-0', task_name='task-2',
                          requirements=requirements)
        requirements = {'max_runtime': 1, 'nodes': 2}
        # This task should be backfilled
        task4 = task.Task(workflow_name='workflow-0', task_name='task-3',
                          requirements=requirements)
        resource1 = resource_allocation.Resource(id_='resource-0', nodes=2)
        resource2 = resource_allocation.Resource(id_='resource-1', nodes=2)
        resource3 = resource_allocation.Resource(id_='resource-2', nodes=2)
        resource4 = resource_allocation.Resource(id_='resource-3', nodes=2)

        tasks = [task1, task2, task3, task4]
        resources = [resource1, resource2, resource3, resource4]
        algorithms.Backfill().schedule_all(tasks, resources)

        assert len(task1.allocations) == 2
        assert all(a.start_time == 0 for a in task1.allocations)
        assert len(task2.allocations) == 4
        start_time = task1.requirements.max_runtime
        assert all(a.start_time == start_time for a in task2.allocations)
        # task3 should not have been backfilled (would have taken too
        # much time)
        assert len(task3.allocations) == 1
        start_time = (task1.requirements.max_runtime
                      + task2.requirements.max_runtime)
        assert task3.allocations[0].start_time == start_time
        # task4 should have been backfilled
        assert len(task4.allocations) == 1
        assert task4.allocations[0].start_time == 0

    @staticmethod
    def test_schedule_six_tasks():
        """Test scheduling six tasks."""
        requirements = {'max_runtime': 1, 'nodes': 1}
        task1 = task.Task(workflow_name='workflow-0', task_name='task-1',
                          requirements=requirements)
        requirements = {'max_runtime': 1, 'nodes': 4}
        task2 = task.Task(workflow_name='workflow-0', task_name='task-2',
                          requirements=requirements)
        requirements = {'max_runtime': 1, 'nodes': 1}
        task3 = task.Task(workflow_name='workflow-0', task_name='task-3',
                          requirements=requirements)
        task4 = task.Task(workflow_name='workflow-0', task_name='task-4',
                          requirements=requirements)
        task5 = task.Task(workflow_name='workflow-0', task_name='task-5',
                          requirements=requirements)
        task6 = task.Task(workflow_name='workflow-0', task_name='task-6',
                          requirements=requirements)
        resource = resource_allocation.Resource(id_='resource-0', nodes=4)

        tasks = [task1, task2, task3, task4, task5, task6]
        algorithms.Backfill().schedule_all(tasks, [resource])

        assert task1.allocations[0].id_ == 'resource-0'
        assert task1.allocations[0].nodes == 1
        assert task1.allocations[0].start_time == 0
        assert task2.allocations[0].id_ == 'resource-0'
        assert task2.allocations[0].nodes == 4
        assert (task2.allocations[0].start_time
                == task1.requirements.max_runtime)
        # Task 3, 4 and 5 should have been backfillled, filling in an area
        # before task 2 can run
        assert task3.allocations[0].id_ == 'resource-0'
        assert task3.allocations[0].nodes == 1
        assert task3.allocations[0].start_time == 0
        assert task4.allocations[0].id_ == 'resource-0'
        assert task4.allocations[0].nodes == 1
        assert task4.allocations[0].start_time == 0
        assert task5.allocations[0].id_ == 'resource-0'
        assert task5.allocations[0].nodes == 1
        assert task5.allocations[0].start_time == 0
        # Task 6 should be scheduled to run last
        assert task6.allocations[0].id_ == 'resource-0'
        assert task6.allocations[0].nodes == 1
        t = task1.requirements.max_runtime + task2.requirements.max_runtime
        assert task6.allocations[0].start_time == t


class TestSJF:
    """Test SJF."""

    @staticmethod
    def test_schedule_one_task():
        """Test scheduling one task."""
        schedule_one_task(algorithms.SJF)

    @staticmethod
    def test_schedule_two_tasks():
        """Test scheduling two tasks."""
        # schedule_two_tasks(algorithms.SJF)
        requirements = {'max_runtime': 3}
        task1 = task.Task(workflow_name='workflow-1', task_name='task-1',
                          requirements=requirements)
        task2 = task.Task(workflow_name='workflow-1', task_name='task-2',
                          requirements=requirements)
        resource = resource_allocation.Resource(id_='test-resource-1', nodes=2)

        algorithms.SJF().schedule_all([task1, task2], [resource])

        assert len(task1.allocations) == 1
        assert task1.allocations[0].id_ == 'test-resource-1'
        assert task1.allocations[0].nodes == 1
        assert len(task2.allocations) == 1
        assert task2.allocations[0].id_ == 'test-resource-1'
        assert task2.allocations[0].nodes == 1
        # These can be allocated at the same time, since the resource has two
        # nodes
        assert (task1.allocations[0].start_time
                == task2.allocations[0].start_time)

    @staticmethod
    def test_schedule_task_fail():
        """Test scheduling a task with more resources required than available."""
        schedule_task_fail(algorithms.SJF())

    @staticmethod
    def test_schedule_task_gpus_req():
        """Test scheduling a task with a gpus requirement."""
        schedule_task_gpus_req(algorithms.SJF())

    @staticmethod
    def test_schedule_task_gpus_req_fail():
        """Test scheduling a task with a gpus requirement that should fail."""
        schedule_task_gpus_req_fail(algorithms.SJF())


#
# Shared testing functions.
#

def schedule_one_task(algorithm):
    """Test scheduling one task."""
    requirements = {'max_runtime': 3}
    task1 = task.Task(workflow_name='workflow-1', task_name='task-1',
                      requirements=requirements)
    resource = resource_allocation.Resource(id_='test-resource-1', nodes=4)

    algorithm.schedule_all([task1], [resource])

    assert task1.allocations[0].id_ == 'test-resource-1'
    assert task1.allocations[0].start_time == 0
    assert task1.allocations[0].nodes == 1


def schedule_two_tasks(algorithm):
    """Test scheduling two tasks."""
    requirements = {'max_runtime': 3}
    task1 = task.Task(workflow_name='workflow-1', task_name='task-1',
                      requirements=requirements)
    task2 = task.Task(workflow_name='workflow-1', task_name='task-2',
                      requirements=requirements)
    resource = resource_allocation.Resource(id_='test-resource-1', nodes=2)

    algorithm.schedule_all([task1, task2], [resource])

    assert task1.allocations[0].id_ == 'test-resource-1'
    assert task1.allocations[0].nodes == 1
    assert task1.allocations[0].start_time == 0
    assert task2.allocations[0].id_ == 'test-resource-1'
    assert task2.allocations[0].nodes == 1
    assert task2.allocations[0].start_time == 0


def schedule_task_fail(algorithm):
    """Test scheduling a task with more resources required than available."""
    requirements = {'max_runtime': 3, 'nodes': 10}
    task1 = task.Task(workflow_name='workflow-1', task_name='task-1',
                      requirements=requirements)
    resource = resource_allocation.Resource(id_='test-resource-1', nodes=2)

    assert algorithm.schedule_all([task1], [resource]) is None
    # No allocations available
    assert not task1.allocations


def schedule_task_gpus_req(algorithm):
    """Test scheduling a task with the gpus requirement."""
    requirements = {
        'max_runtime': 10,
        'nodes': 10,
        'gpus_per_node': 4,
    }
    task1 = task.Task(workflow_name='test-workflow', task_name='task-1',
                      requirements=requirements)
    resource1 = resource_allocation.Resource(id_='test-resource-1', nodes=20,
                                             gpus_per_node=0)
    resource2 = resource_allocation.Resource(id_='test-resource-2', nodes=20,
                                             gpus_per_node=20)

    algorithm.schedule_all([task1], [resource1, resource2])

    assert task1.allocations[0].id_ == 'test-resource-2'
    assert task1.allocations[0].nodes == 10
    assert task1.allocations[0].start_time == 0


def schedule_task_gpus_req_fail(algorithm):
    """Test scheduling a task with the gpus requirement that should fail."""
    requirements = {
        'max_runtime': 10,
        'nodes': 10,
        'gpus_per_node': 10,
    }
    task1 = task.Task(workflow_name='test-workflow', task_name='task-1',
                      requirements=requirements)
    resource1 = resource_allocation.Resource(id_='test-resource-1', nodes=20,
                                             gpus_per_node=0)

    algorithm.schedule_all([task1], [resource1])

    assert not task1.allocations



#!/usr/bin/env python3
"""REST Interface for the BEE Scheduler."""

import argparse
import os

from flask import Flask, request
from flask_restful import Resource, Api

from beeflow.scheduler import algorithms
from beeflow.scheduler import task
from beeflow.scheduler import resource_allocation
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.db import sched
from beeflow.common import log as bee_logging


log = bee_logging.setup(__name__)

flask_app = Flask(__name__)
api = Api(flask_app)

# We have to call bc.init() here due to how gunicorn works
bc.init()


def get_db_path():
    """Get the database path."""
    # Favor the environment variable if it exists
    path = os.getenv('BEE_SCHED_DB_PATH')
    if path is None:
        workdir = bc.get('DEFAULT', 'bee_workdir')
        path = os.path.join(workdir, 'sched.db')
    return path


def connect_db(fn):
    """Decorate a function for connecting to a database."""
    def wrap(*pargs, **kwargs):
        """Decorate the function."""
        with sched.open_db(get_db_path()) as db:
            return fn(db, *pargs, **kwargs)

    return wrap


class ResourcesHandler(Resource):
    """Resources handler."""

    @staticmethod
    @connect_db
    def put(db):
        """Create a list of resources to use for allocation."""
        db.resources.clear()
        resources = [resource_allocation.Resource.decode(r)
                     for r in request.json]
        db.resources.extend(resources)
        return f'created {len(resources)} resource(s)'

    @staticmethod
    @connect_db
    def get(db):
        """Get a list of all resources."""
        return [r.encode() for r in db.resources]


class WorkflowJobHandler(Resource):
    """Schedule jobs for a specific workflow with the current resources."""

    @staticmethod
    @connect_db
    def put(db, workflow_name):  # noqa ('workflow_name' may be used in the future)
        """Schedules a new list of independent tasks with available resources."""
        data = request.json
        tasks = [task.Task.decode(t) for t in data]
        # Pick the scheduling algorithm
        algorithm = algorithms.choose(**vars(flask_app.sched_conf))
        algorithm.schedule_all(tasks, list(db.resources))
        return [t.encode() for t in tasks]


api.add_resource(ResourcesHandler, '/bee_sched/v1/resources')
api.add_resource(WorkflowJobHandler,
                 '/bee_sched/v1/workflows/<string:workflow_name>/jobs')

# Default config values
SCHEDULER_PORT = 5100
# TODO: Use MODEL_FILE when interacting with MARS scheduling
MODEL_FILE = 'model'
ALLOC_LOGFILE = 'schedule_trace.txt'
MARS_CNT = 4
DEFAULT_ALGORITHM = 'fcfs'


def load_config_values():
    """Load the config, if necessary, and return config values.

    Load the config, if necessary, and return config values.
    """
    # Set the default config values
    conf = {
        'log': None,
        'alloc_logfile': None,
        'algorithm': None,
        'default_algorithm': None,
        'workdir': None,
    }

    for key in conf:
        conf[key] = bc.get('scheduler', key)
    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    # Set some defaults
    if not conf['log']:
        conf['log'] = '/'.join([bee_workdir, 'logs', 'scheduler.log'])
    if not conf['workdir']:
        conf['workdir'] = os.path.join(bee_workdir, 'scheduler')
    if not conf['alloc_logfile']:
        conf['alloc_logfile'] = os.path.join(conf['workdir'],
                                             ALLOC_LOGFILE)

    conf = argparse.Namespace(**conf)
    log.info('Config = [')
    log.info(f'\talloc_logfile = {conf.alloc_logfile}')
    log.info(f'\talgorithm = {conf.algorithm}')
    log.info(f'\tdefault_algorithm = {conf.default_algorithm}')
    log.info(f'\tworkdir = {conf.workdir}')
    log.info(']')
    return conf


def create_app():
    """Create the Flask app for the scheduler."""
    # TODO: Refactor this to actually create the app here
    conf = load_config_values()
    flask_app.sched_conf = conf
    # Load algorithm data
    algorithms.load(**vars(conf))

    # Create the scheduler workdir, if necessary
    # sched_listen_port = wf_utils.get_open_port()
    # wf_db.set_sched_port(sched_listen_port)
    os.makedirs(conf.workdir, exist_ok=True)
    return flask_app

# Ignore todo's or pylama fails
# pylama:ignore=W0511



"""Mid-level interface for container runtime system.

Delegates the writing of the text for job script to an instance of a subclass
of the abstract base class 'ContainerRuntimeDriver'.
Default: 'CharliecloudDriver' class.
"""

from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.crt.charliecloud_driver import CharliecloudDriver
from beeflow.common.crt.singularity_driver import SingularityDriver


class ContainerRuntimeInterface:
    """Interface for the container runtime.

    Requires an implemented subclass of ContainerRuntimeDriver to function.
    """

    def __init__(self, crt_driver=CharliecloudDriver):
        """Initialize the CRT interface with a runtime, CharliecloudDriver by default.

        :param crt_driver: container runtime driver (default: CharliecloudDriver)
        :type crt_driver: subclass of ContainerRuntimeDriver
        """
        self._crt_driver = crt_driver()

    def run_text(self, task):
        """Create text required to run the task using the container_runtime.

        :param task: instance of Task
        :rtype: string
        """
        return self._crt_driver.run_text(task)

    def build_text(self, userconfig, task):
        """Create text required to build a task environment.

        :param task: instance of Task
        :param userconfig: path to userconfig file
        :rtype: string
        """
        return self._crt_driver.build_text(userconfig, task)
# Ignore module imported but unused error. No way to know which crt will be needed
# pylama:ignore=W0611



"""Contains functions for managing a database for workflow and task information."""

import sqlite3
from sqlite3 import Error
from collections import namedtuple
import os
from beeflow.common.config_driver import BeeConfig as bc


def create_connection():
    """Create a new connection with the workflow database."""
    db_file = os.path.join(bc.get('DEFAULT', 'bee_workdir'), 'workflow.db')
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Error as error:
        print(error)
    return conn


def create_table(stmt):
    """Create a new table in the database."""
    with create_connection() as conn:
        try:
            cursor = conn.cursor()
            cursor.execute(stmt)
        except Error as error:
            print(error)


def run(stmt, params=None):
    """Run the sql statement on the database. Doesn't return anything."""
    with create_connection() as conn:
        try:
            cursor = conn.cursor()
            if params:
                cursor.execute(stmt, params)
            else:
                cursor.execute(stmt)
            conn.commit()
        except Error as error:
            print(error)


def get(stmt, params=None):
    """Run the sql statement on the database and return the result."""
    with create_connection() as conn:
        try:
            cursor = conn.cursor()
            if params:
                cursor.execute(stmt, params)
            else:
                cursor.execute(stmt)
            result = cursor.fetchall()
        except Error:
            result = None
        return result


def table_exists(table_name):
    """Return true if a table exists and false if not."""
    stmt = f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';"
    result = get(stmt)
    return len(result) != 0


def get_table_length(table):
    """Return the number of rows in a table."""
    stmt = f"SELECT COUNT(*) from {table}"
    result = get(stmt)
    rows = result[0][0]
    return rows


def init_info():
    """Insert a new workflow into the database."""
    stmt = """INSERT INTO info (wfm_port, tm_port, sched_port, num_workflows)
                                    VALUES(?, ?, ?, ?);"""
    run(stmt, [-1, -1, -1, 0])


def init_tables():
    """Create the database."""
    # Create tables if they don't exist
    # Create new database
    workflows_stmt = """CREATE TABLE IF NOT EXISTS workflows (
                            id INTEGER PRIMARY KEY,
                            -- Set workflow ID to unique.
                            workflow_id INTEGER UNIQUE,
                            name TEXT,
                            status TEST NOT NULL,
                            run_dir STR,
                            bolt_port INTEGER,
                            gdb_pid INTEGER);"""

    tasks_stmt = """CREATE TABLE IF NOT EXISTS tasks (
                    id INTEGER PRIMARY KEY,
                    task_id INTEGER UNIQUE,
                    workflow_id INTEGER NOT NULL,
                    name TEXT,
                    resource TEXT,
                    status TEXT,
                    slurm_id INTEGER,
                    FOREIGN KEY (workflow_id)
                        REFERENCES workflows (workflow_id)
                            ON DELETE CASCADE
                            ON UPDATE NO ACTION);"""

    info_stmt = """CREATE TABLE IF NOT EXISTS info (
                       id INTEGER PRIMARY KEY,
                       wfm_port INTEGER,
                       tm_port INTEGER,
                       sched_port INTEGER,
                       num_workflows INTEGER
                       );"""

    if not table_exists('workflows'):
        create_table(workflows_stmt)
    if not table_exists('tasks'):
        create_table(tasks_stmt)
    if not table_exists('info'):
        create_table(info_stmt)
        init_info()


Info = namedtuple("Info", "id wfm_port tm_port sched_port num_workflows")


def get_info():
    """Return an info object containing port information."""
    stmt = "SELECT * FROM info"
    result = get(stmt)
    info = Info(*result[0])
    return info


def get_wfm_port():
    """Return workflow manager port."""
    stmt = "SELECT wfm_port FROM info"
    result = get(stmt)
    wfm_port = result[0][0]
    return wfm_port


def get_tm_port():
    """Return task manager port."""
    stmt = "SELECT tm_port FROM info"
    result = get(stmt)
    tm_port = result[0][0]
    return tm_port


def get_sched_port():
    """Return scheduler port."""
    stmt = "SELECT sched_port FROM info"
    result = get(stmt)
    sched_port = result[0][0]
    return sched_port


def get_num_workflows():
    """Return scheduler port."""
    stmt = "SELECT num_workflows FROM info"
    result = get(stmt)
    sched_port = result[0][0]
    return sched_port


def increment_num_workflows():
    """Set workflow manager port."""
    stmt = "UPDATE info SET num_workflows = num_workflows + 1"
    run(stmt)


def set_wfm_port(new_port):
    """Set workflow manager port."""
    if not table_exists('info'):
        # Initialize the database
        init_tables()

    stmt = "UPDATE info SET wfm_port=?"
    run(stmt, [new_port])


def set_tm_port(new_port):
    """Set workflow manager port."""
    if not table_exists('info'):
        # Initialize the database
        init_tables()
    stmt = "UPDATE info SET tm_port=?"
    run(stmt, [new_port])


def set_sched_port(new_port):
    """Set workflow manager port."""
    if not table_exists('info'):
        # Initialize the database
        init_tables()
    stmt = "UPDATE info SET sched_port=?"
    run(stmt, [new_port])


def add_workflow(workflow_id, name, status, run_dir, bolt_port, gdb_pid):
    """Insert a new workflow into the database."""
    if not table_exists('workflows'):
        # Initialize the database
        init_tables()

    stmt = ("INSERT INTO workflows (workflow_id, name, status, run_dir, bolt_port, gdb_pid) "
            "VALUES(?, ?, ?, ?, ?, ?);"
            )
    run(stmt, [workflow_id, name, status, run_dir, bolt_port, gdb_pid])


def update_workflow_state(workflow_id, status):
    """Update the status in a workflow in the database."""
    stmt = "UPDATE workflows SET status=? WHERE workflow_id=?"
    run(stmt, [status, workflow_id])


Workflow = namedtuple("Workflow", "id workflow_id name status run_dir bolt_port gdb_pid")


def get_workflow_state(workflow_id):
    """Return the bolt port associated with a workflow."""
    stmt = "SELECT state FROM workflows WHERE workflow_id=?"
    result = get(stmt, [workflow_id])[0]
    state = result[0]
    return state


def get_bolt_port(workflow_id):
    """Return the bolt port associated with a workflow."""
    stmt = "SELECT bolt_port FROM workflows WHERE workflow_id=?"
    result = get(stmt, [workflow_id])
    result = get(stmt, [workflow_id])[0]
    bolt_port = result[0]
    return bolt_port


def get_gdb_pid(workflow_id):
    """Return the bolt port associated with a workflow."""
    stmt = "SELECT gdb_pid FROM workflows WHERE workflow_id=?"
    result = get(stmt, [workflow_id])[0]
    gdb_pid = result[0]
    return gdb_pid


def get_run_dir(workflow_id):
    """Return the bolt port associated with a workflow."""
    stmt = "SELECT run_dir FROM workflows WHERE workflow_id=?"
    result = get(stmt, [workflow_id])[0]
    run_dir = result[0]
    return run_dir


def get_workflow(workflow_id):
    """Return a workflow object."""
    stmt = "SELECT * FROM workflows WHERE workflow_id=?"
    result = get(stmt, [workflow_id])[0]
    workflow = Workflow(*result)
    return workflow


def get_workflows():
    """Return a list of all the workflows."""
    stmt = "SELECT * FROM workflows"
    result = get(stmt)
    workflows = [Workflow(*workflow) for workflow in result]
    return workflows


def add_task(task_id, workflow_id, name, status):
    """Add a task to the database associated with the workflow id specified."""
    stmt = "INSERT INTO tasks (task_id, workflow_id, name, resource, status,"\
           "slurm_id) VALUES(?, ?, ?, ?, ?, ?)"
    run(stmt, [task_id, workflow_id, name, "", status, -1])


Task = namedtuple("Task", "id task_id workflow_id name resource status"
                  " slurm_id")


def get_tasks(workflow_id):
    """Get all tasks associated with a particular workflow."""
    stmt = "SELECT * FROM tasks WHERE workflow_id=?"
    result = get(stmt, [workflow_id])
    tasks = [Task(*task) for task in result]
    return tasks


def delete_workflow(workflow_id):
    """Delete a workflow from the database."""
    stmt = "DELETE FROM workflows WHERE workflow_id=?"
    run(stmt, [workflow_id])


def update_task_state(task_id, workflow_id, status):
    """Update the state of a task."""
    stmt = "UPDATE tasks SET status=? WHERE task_id=? AND workflow_id=? "
    run(stmt, [status, task_id, workflow_id])


def get_task(task_id, workflow_id):
    """Get a task associed with a workflow."""
    stmt = "SELECT * FROM task WHERE task_id=? AND workflow_id=?"
    result = get(stmt, [task_id, workflow_id])
    return result


def delete_task(task_id, workflow_id):
    """Delete a task associed with a workflow."""
    stmt = "DELETE FROM tasks WHERE task_id=? AND workflow_id=?"
    run(stmt, [task_id, workflow_id])



"""Code for CWL expressions."""

import re


def eval_output(input_pairs, glob):
    """Evaluate a simple output expression."""
    # Find all matches of string pattern $(inputs.foo), capturing value of foo
    expr_pattern = r"\$\(inputs\.(\w+)\)"
    match = re.findall(expr_pattern, glob)
    if match:
        # Split string up to get list of non-matching characters
        split_pattern = r"\$\(inputs\.\w+\)"
        split = re.split(split_pattern, glob)
        values = []
        for m in match:
            print(m)
            if m in input_pairs.keys():
                values.append(input_pairs[m])
            else:
                raise ValueError(f"reference to non-existent task input {m}")
        # Construct string with evaluated inputs
        eval_string = split[0]
        for v, s in zip(values, split[1:]):
            eval_string += v + s
        return eval_string
    raise RuntimeError(f'could not parse output expression in glob: {glob}')


def eval_input(input_pairs, value_from):
    """Evaluate a simple input expression."""
    # Match any of the following patterns: self.path, inputs.foo, or "foo"
    # in an expression of the form $(A + B) in the valueFrom field
    expr_pattern = r'\$\((self\.path|inputs\.\w+|".+") \+ (self\.path|inputs\.\w+|".+")\)'
    match = re.fullmatch(expr_pattern, value_from)
    if match:
        values = []
        for m in match.groups():
            inputs_dot = "inputs."
            if m == "self.path":
                raise NotImplementedError('Using "self.path" in a "value_from" field '
                                          'is not yet supported')
            if m.startswith(inputs_dot):
                if m[len(inputs_dot):] in input_pairs.keys():
                    values.append(input_pairs[m[len(inputs_dot):]])
                else:
                    raise ValueError(f"reference to non-existent task input {m[len(inputs_dot):]}")
            else:
                # Pattern "foo" (strip quotes)
                values.append(m[1:-1])
        return values[0] + values[1]
    raise ValueError(f"unable to evaluate expression {value_from}")



"""The workflow list module.

This contains endpoints forsubmitting, starting, and reexecuting workflows.
"""

import os
import subprocess
import traceback
import jsonpickle

from flask import make_response, jsonify
from werkzeug.datastructures import FileStorage
from flask_restful import Resource, reqparse

from beeflow.common import log as bee_logging
# from beeflow.common.wf_profiler import WorkflowProfiler
from beeflow.common.parser import CwlParser, CwlParseError

from beeflow.wf_manager.resources import wf_utils
from beeflow.wf_manager.common import dep_manager
from beeflow.wf_manager.common import wf_db
from beeflow.common import wf_data

log = bee_logging.setup(__name__)


def parse_workflow(wfi, wf_id, workflow_dir, main_cwl, yaml_file):
    """Run the parser."""
    parser = CwlParser(wfi)
    cwl_path = os.path.join(workflow_dir, main_cwl)
    if yaml_file is not None:
        yaml_path = os.path.join(workflow_dir, yaml_file)
        parser.parse_workflow(wf_id, cwl_path, yaml_path)
    else:
        parser.parse_workflow(wf_id, cwl_path)


# def initialize_wf_profiler(wf_name):
#    # Initialize the workflow profiling code
#    bee_workdir = wf_utils.get_bee_workdir()
#    fname = '{}.json'.format(wf_name)
#    profile_dir = os.path.join(bee_workdir, 'profiles')
#    os.makedirs(profile_dir, exist_ok=True)
#    output_path = os.path.join(profile_dir, fname)
#    wf_profiler = WorkflowProfiler(wf_name, output_path)


def extract_wf(wf_id, filename, workflow_archive, reexecute=False):
    """Extract a workflow into the workflow directory."""
    wf_utils.create_workflow_dir(wf_id)
    wf_dir = wf_utils.get_workflow_dir(wf_id)
    archive_path = os.path.join(wf_dir, filename)
    workflow_archive.save(archive_path)
    if not reexecute:
        subprocess.run(['tar', '-xf', archive_path, '-C', wf_dir], check=False)
        cwl_dir = os.path.join(wf_dir, filename.split('.')[0])
        return cwl_dir

    subprocess.run(['tar', '-xf', archive_path, '--strip-components=1',
                    '-C', wf_dir], check=False)
    archive_dir = os.path.join(wf_dir, 'gdb')
    return archive_dir


class WFList(Resource):
    """Interacts with existing workflows."""

    def get(self):
        """Return list of workflows to client."""
        workflow_list = wf_db.get_workflows()
        info = []
        for wf_info in workflow_list:
            wf_id = wf_info.workflow_id
            wf_status = wf_info.status
            wf_name = wf_info.name
            info.append([wf_name, wf_id, wf_status])
        resp = make_response(jsonify(workflow_list=jsonpickle.encode(info)), 200)
        return resp

    def post(self):
        """Receive a workflow, parse it, and start up a neo4j instance for it."""
        reqparser = reqparse.RequestParser()
        reqparser.add_argument('wf_name', type=str, required=True,
                               location='form')
        reqparser.add_argument('main_cwl', type=str, required=True,
                               location='form')
        reqparser.add_argument('yaml', type=str, required=False,
                               location='form')
        reqparser.add_argument('wf_filename', type=str, required=True,
                               location='form')
        reqparser.add_argument('workdir', type=str, required=True,
                               location='form')
        reqparser.add_argument('workflow_archive', type=FileStorage, required=False,
                               location='files')
        data = reqparser.parse_args()
        wf_tarball = data['workflow_archive']
        wf_filename = data['wf_filename']
        main_cwl = data['main_cwl']
        wf_name = data['wf_name']
        wf_workdir = data['workdir']
        # None if not sent
        yaml_file = data['yaml']

        try:
            dep_manager.create_image()
        except dep_manager.NoContainerRuntime:
            crt_message = "Charliecloud not installed in current environment."
            log.error(crt_message)
            resp = make_response(jsonify(msg=crt_message, status='error'), 418)
            return resp

        wf_id = wf_data.generate_workflow_id()
        wf_dir = extract_wf(wf_id, wf_filename, wf_tarball)
        bolt_port = wf_utils.get_open_port()
        http_port = wf_utils.get_open_port()
        https_port = wf_utils.get_open_port()
        gdb_pid = dep_manager.start_gdb(wf_dir, bolt_port, http_port, https_port)
        wf_db.add_workflow(wf_id, wf_name, 'Pending', wf_dir, bolt_port, gdb_pid)
        dep_manager.wait_gdb(log)

        try:
            # wfi = parse_workflow(wf_path, main_cwl, yaml_file)
            wfi = wf_utils.get_workflow_interface(wf_id)
            parse_workflow(wfi, wf_id, wf_dir, main_cwl, yaml_file)
        except CwlParseError as err:
            traceback.print_exc()
            log.error('Failed to parse file')
            return make_response(jsonify(msg=f'Parser: {err.args[0]}', status='error'), 400)
        # initialize_wf_profiler(wf_name)

        wf_utils.create_wf_metadata(wf_id, wf_name)
        _, tasks = wfi.get_workflow()
        for task in tasks:
            metadata = wfi.get_task_metadata(task)
            metadata['workdir'] = wf_workdir
            wfi.set_task_metadata(task, metadata)
            wf_db.add_task(task.id, wf_id, task.name, "WAITING")
        resp = make_response(jsonify(msg='Workflow uploaded', status='ok',
                             wf_id=wf_id), 201)
        return resp

    def put(self):
        """Reexecute a workflow."""
        reqparser = reqparse.RequestParser()
        reqparser.add_argument('wf_name', type=str, required=True,
                               location='form')
        reqparser.add_argument('wf_filename', type=str, required=True,
                               location='form')
        reqparser.add_argument('workdir', type=str, required=True,
                               location='form')
        reqparser.add_argument('workflow_archive', type=FileStorage, required=False,
                               location='files')
        data = reqparser.parse_args()
        workflow_archive = data['workflow_archive']
        wf_filename = data['wf_filename']
        wf_name = data['wf_name']
        wf_workdir = data['workdir']

        try:
            dep_manager.create_image()
        except dep_manager.NoContainerRuntime:
            crt_message = "Charliecloud not installed in current environment."
            log.error(crt_message)
            resp = make_response(jsonify(msg=crt_message, status='error'), 418)
            return resp

        wf_id = wf_data.generate_workflow_id()
        wf_dir = extract_wf(wf_id, wf_filename, workflow_archive, reexecute=True)
        bolt_port = wf_utils.get_open_port()
        http_port = wf_utils.get_open_port()
        https_port = wf_utils.get_open_port()
        gdb_pid = dep_manager.start_gdb(wf_dir, bolt_port, http_port,
                                        https_port, reexecute=True)
        wf_db.add_workflow(wf_id, wf_name, 'Pending', wf_dir, bolt_port, gdb_pid)
        dep_manager.wait_gdb(log)
        wfi = wf_utils.get_workflow_interface(wf_id)
        wfi.reset_workflow(wf_id)
        wf_utils.create_wf_metadata(wf_id, wf_name)

        _, tasks = wfi.get_workflow()
        for task in tasks:
            metadata = wfi.get_task_metadata(task)
            metadata['workdir'] = wf_workdir
            wfi.set_task_metadata(task, metadata)
        # Return the wf_id and created
        resp = make_response(jsonify(msg='Workflow uploaded', status='ok',
                             wf_id=wf_id), 201)
        return resp

    def patch(self):
        """Copy workflow archive."""
        reqparser = reqparse.RequestParser()
        data = reqparser.parse_args()
        bee_workdir = wf_utils.get_bee_workdir()
        wf_id = data['wf_id']
        archive_path = os.path.join(bee_workdir, 'archives', wf_id + '.tgz')
        with open(archive_path, 'rb') as archive:
            archive_file = jsonpickle.encode(archive.read())
        archive_filename = os.path.basename(archive_path)
        resp = make_response(jsonify(archive_file=archive_file,
                             archive_filename=archive_filename), 200)
        return resp



"""Connection class for connecting to other components over a socket."""
import urllib
import os
import requests

import requests_unixsocket


class Connection:
    """Connection for sending/receiving requests from a component."""

    def __init__(self, socket, prefix=None, error_handler=None):
        """Construct a new connection from a socket path."""
        self._socket_path = urllib.parse.quote(socket, safe='')
        self._session = requests_unixsocket.Session()
        self._prefix = prefix
        self._error_handler = (error_handler if error_handler is not None
                               else lambda resp: resp)

    def _full_url(self, path):
        """Get the full url for a path."""
        # For some reason urllib.parse.urljoin() is just returning `path` here
        # return urllib.parse.urljoin('http+unix://{self._socket_path}', path)
        if self._prefix is not None:
            path = os.path.join(self._prefix, path)
        return os.path.join(f'http+unix://{self._socket_path}', path)

    def handle_error(self, resp):
        """Handle an error, if there is one."""
        if resp.status_code != requests.codes.okay:  # noqa (pylama can't find the okay member)
            return self._error_handler(resp)
        return resp

    def get(self, path, *pargs, **kwargs):
        """Do an HTTP GET request."""
        return self.handle_error(self._session.get(self._full_url(path),
                                                   *pargs, **kwargs))

    def put(self, path, *pargs, **kwargs):
        """Do an HTTP PUT request."""
        return self.handle_error(self._session.put(self._full_url(path),
                                                   *pargs, **kwargs))

    def post(self, path, *pargs, **kwargs):
        """Do an HTTP POST request."""
        return self.handle_error(self._session.post(self._full_url(path),
                                                    *pargs, **kwargs))

    def delete(self, path, *pargs, **kwargs):
        """Do an HTTP DELETE request."""
        return self.handle_error(self._session.delete(self._full_url(path),
                                                      *pargs, **kwargs))

    def patch(self, path, *pargs, **kwargs):
        """Do an HTTP PATCH request."""
        return self.handle_error(self._session.patch(self._full_url(path),
                                                     *pargs, **kwargs))



#! /usr/bin/env python3
"""Unit test module for the BEE CWL parser module."""

from pathlib import Path
import unittest
from beeflow.common.parser import CwlParser
from beeflow.common.wf_data import generate_workflow_id
from beeflow.tests.mocks import MockWFI

# Disable protected member access warning
# pylama:ignore=W0212

REPO_PATH = Path(*Path(__file__).parts[:-3])


def find(path):
    """Find a path relative to the root of the repo."""
    return str(Path(REPO_PATH, path))


class TestParser(unittest.TestCase):
    """Unit test case for the workflow interface."""

    @classmethod
    def setUpClass(cls):
        """Start the GDB, initialize the CWL parser, which connects to the GDB."""
        cls.wfi = MockWFI()
        cls.parser = CwlParser(cls.wfi)

    @classmethod
    def tearDownClass(cls):
        """Stop the GDB."""

    def tearDown(self):
        """Clear all data in the Neo4j database."""
        if self.wfi.workflow_initialized() and self.wfi.workflow_loaded():
            self.wfi.finalize_workflow()

    def test_parse_workflow(self):
        """Test parsing of workflow with an input job file."""
        cwl_wfi_file = find("examples/clamr-ffmpeg-build/clamr_wf.cwl")
        cwl_job_yaml = find("examples/clamr-ffmpeg-build/clamr_job.yml")
        cwl_job_json = find("examples/clamr-ffmpeg-build/clamr_job.json")
        workflow_id = generate_workflow_id()

        # Test workflow parsing with YAML input job file
        wfi = self.parser.parse_workflow(workflow_id, cwl_wfi_file, cwl_job_yaml)
        self.assertTrue(wfi.workflow_loaded())

        wfi.finalize_workflow()
        self.assertFalse(wfi.workflow_loaded())

        # Test workflow parsing with JSON input job file
        wfi = self.parser.parse_workflow(workflow_id, cwl_wfi_file, cwl_job_json)
        self.assertTrue(wfi.workflow_loaded())

    def test_parse_workflow_no_job(self):
        """Test parsing of a workflow without an input job file."""
        cwl_wfi_file = find("beeflow/tests/cf.cwl")
        workflow_id = generate_workflow_id()
        # cwl_wfi_file = "examples/clamr-ffmpeg-build/clamr_wf.cwl"

        # Test workflow parsing without input job file
        wfi = self.parser.parse_workflow(workflow_id, cwl_wfi_file)
        self.assertTrue(wfi.workflow_loaded())


if __name__ == '__main__':
    unittest.main()



"""Tests for container path conversion."""

from beeflow.common.container_path import convert_path


def test_empty():
    """Test an empty bind mounts list."""
    assert convert_path('/home/test', {}) == '/home/test'
    assert convert_path('/usr', {}) == '/usr'


def test_nonempty():
    """Test a non-empty bind mount list."""
    bind_mounts = {
        '/prefix/home/test': '/home/test',
        '/usr/projects/some-project': '/mnt/0',
        '/data': '/mnt/1',
    }

    assert convert_path('/prefix/home/test/some/path', bind_mounts) == '/home/test/some/path'
    assert convert_path('/usr/projects/some-project/123', bind_mounts) == '/mnt/0/123'
    assert convert_path('/data/1/2/3', bind_mounts) == '/mnt/1/1/2/3'
    assert convert_path('/some/other/path', bind_mounts) == '/some/other/path'



"""Slurm worker for work load management.

Builds command for submitting batch job.
"""

import subprocess
import json
import urllib
import getpass
import requests_unixsocket
import requests

from beeflow.common import log as bee_logging
from beeflow.common.worker.worker import (Worker, WorkerError)

log = bee_logging.setup(__name__)


class SlurmWorker(Worker):
    """The Worker for systems where Slurm is the Work Load Manager."""

    def __init__(self, bee_workdir, **kwargs):
        """Create a new Slurm Worker object."""
        super().__init__(bee_workdir, **kwargs)
        # Pull slurm socket configs from kwargs (Uses getpass.getuser() instead
        # of os.getlogin() because of an issue with using getlogin() without a
        # controlling terminal)
        self.slurm_socket = kwargs.get('slurm_socket', f'/tmp/slurm_{getpass.getuser()}.sock')
        self.session = requests_unixsocket.Session()
        encoded_path = urllib.parse.quote(self.slurm_socket, safe="")
        # Note: Socket path is encoded, http request is not generally.
        self.slurm_url = f"http+unix://{encoded_path}/slurm/v0.0.35"

    def build_text(self, task):
        """Build text for task script; use template if it exists."""
        task_save_path = self.task_save_path(task)
        crt_res = self.crt.run_text(task)
        requirements = dict(task.requirements)
        hints = dict(task.hints)
        main_command = crt_res.main_command[:]
        stdout_param = ['--output', task.stdout]
        stderr_param = ['--error', task.stderr]
        if task.stdout and task.stderr:
            main_command = stdout_param + stderr_param + main_command
        elif task.stdout:
            main_command = stdout_param + main_command
        elif task.stderr:
            main_command = stderr_param + main_command

        job_text = self.template.render(
            task_save_path=task_save_path,
            task_name=task.name,
            task_id=task.id,
            workflow_id=task.workflow_id,
            env_code=crt_res.env_code,
            pre_commands=crt_res.pre_commands,
            main_command=main_command,
            post_commands=crt_res.post_commands,
            requirements=requirements,
            hints=hints,
        )
        return job_text

    def write_script(self, task):
        """Build task script; returns filename of script."""
        task_text = self.build_text(task)

        task_script = f'{self.task_save_path(task)}/{task.name}-{task.id}.sh'
        with open(task_script, 'w', encoding='UTF-8') as script_f:
            script_f.write(task_text)
            script_f.close()
        return task_script

    @staticmethod
    def query_job(job_id, session, slurm_url):
        """Query slurm for job status."""
        try:
            resp = session.get(f'{slurm_url}/job/{job_id}')

            if resp.status_code != 200:
                raise WorkerError(f'Failed to query job {job_id}')
            status = json.loads(resp.text)
            # For some versions of slurm, the job_state isn't included on failure
            try:
                job_state = status['job_state']
            except KeyError as exc:
                raise WorkerError(f'Failed to query job {job_id}') from exc
        except requests.exceptions.ConnectionError:
            job_state = "NOT_RESPONDING"
        return job_state

    def submit_job(self, script, session, slurm_url):
        """Worker submits job-returns (job_id, job_state)."""
        job_st = subprocess.check_output(['sbatch', '--parsable', script],
                                         stderr=subprocess.STDOUT)
        job_id = int(job_st)
        job_state = self.query_job(job_id, session, slurm_url)
        return job_id, job_state

    def submit_task(self, task):
        """Worker builds & submits script."""
        task_script = self.write_script(task)
        job_id, job_state = self.submit_job(task_script, self.session, self.slurm_url)
        return job_id, job_state

    def query_task(self, job_id):
        """Worker queries job; returns job_state."""
        job_state = self.query_job(job_id, self.session, self.slurm_url)
        return job_state

    def cancel_task(self, job_id):
        """Worker cancels job, returns job_state."""
        try:
            resp = self.session.delete(f'{self.slurm_url}/job/{job_id}')
        except requests.exceptions.ConnectionError:
            return 'NOT_RESPONDING'
        # For some reason, some versions of slurmrestd are not returning an
        # HTTP error code, but just an error in the body
        err = WorkerError(f'Unable to cancel job id {job_id}!')
        if resp.status_code != 200:
            raise err
        try:
            data = resp.json()
            if 'error' in data:
                raise err
        except requests.exceptions.JSONDecodeError as exc:  # noqa requests is not installed in CI
            raise err from exc
        job_state = "CANCELLED"
        return job_state



"""Task Manager submits & manages tasks from Work Flow Manager.

Submits, cancels and monitors states of tasks.
Communicates status to the Work Flow Manager, through RESTful API.
"""
import atexit
import sys
import hashlib
import os
from pathlib import Path
import re
import socket
import traceback
import jsonpickle

from flask import Flask, jsonify, make_response
from flask_restful import Resource, Api, reqparse

from apscheduler.schedulers.background import BackgroundScheduler

from beeflow.common.config_driver import BeeConfig as bc

# This must be imported before calling other parts of BEE
bc.init()

print(sys.argv)

from beeflow.common import log as bee_logging
from beeflow.common.build_interfaces import build_main
from beeflow.common.worker_interface import WorkerInterface
from beeflow.common.connection import Connection
import beeflow.common.worker as worker_pkg
from beeflow.common.db import tm


log = bee_logging.setup(__name__)

runtime = bc.get('task_manager', 'container_runtime')


flask_app = Flask(__name__)
api = Api(flask_app)

# submit_queue = []  # tasks ready to be submitted
# job_queue = []  # jobs that are being monitored
DB_PATH = os.path.join(bc.get('DEFAULT', 'bee_workdir'), 'tm.db')


def connect_db(fn):
    """Connect to the TM database."""

    def wrap(*pargs, **kwargs):
        """Wrap the function."""
        # Check for the TESTING_DB_PATH for running the unit tests
        try:
            db_path = flask_app.config['TESTING_DB_PATH']
        except KeyError:
            db_path = DB_PATH
        with tm.open_db(db_path) as db:
            return fn(db, *pargs, **kwargs)

    return wrap


def _url():
    """Return  the url to the WFM."""
    # Saving this for whenever we need to run jobs across different machines
    # workflow_manager = 'bee_wfm/v1/jobs/'
    # #wfm_listen_port = bc.get('workflow_manager', 'listen_port')
    # wfm_listen_port = wf_db.get_wfm_port()
    # return f'http://127.0.0.1:{wfm_listen_port}/{workflow_manager}'
    return 'bee_wfm/v1/jobs/'


def _resource(tag=""):
    """Access the WFM."""
    return _url() + str(tag)


def _wfm_conn():
    """Get a new connection to the WFM."""
    return Connection(bc.get('workflow_manager', 'socket'))


def update_task_state(workflow_id, task_id, job_state, **kwargs):
    """Informs the workflow manager of the current state of a task."""
    data = {'wf_id': workflow_id, 'task_id': task_id, 'job_state': job_state}
    if 'metadata' in kwargs:
        kwargs['metadata'] = jsonpickle.encode(kwargs['metadata'])

    if 'task_info' in kwargs:
        kwargs['task_info'] = jsonpickle.encode(kwargs['task_info'])

    data.update(kwargs)
    conn = _wfm_conn()
    resp = conn.put(_resource("update/"), json=data)
    if resp.status_code != 200:
        log.warning("WFM not responding when sending task update.")


def update_task_metadata(task_id, metadata):
    """Send workflow manager task metadata."""
    log.info(f'Update task metadata for {task_id}:\n {metadata}')
    # resp = requests.put(_resource("update/"), json=metadata)
    # if resp.status_code != 200:
    #     log.warning("WFM not responding when sending task metadata.")


def gen_task_metadata(task, job_id):
    """Generate dictionary of task metadata for the job submitted.

    Includes:
       job_id
       hostname
       container runtime (when task uses a container)
       hash of container file (when task uses a container)
    """
    hostname = socket.gethostname()
    metadata = {'job_id': job_id, 'host': hostname}
    for hint in task.hints:
        if hint.class_ == "DockerRequirement" and "dockerImageId" in hint.params.keys():
            metadata['container_runtime'] = bc.get('task_manager', 'container_runtime')
            container_path = hint.params["dockerImageId"]
            with open(container_path, 'rb') as container:
                c_hash = hashlib.md5()
                chunk = container.read(8192)
                while chunk:
                    c_hash.update(chunk)
                    chunk = container.read(8192)
            container_hash = c_hash.hexdigest()
            metadata['container_hash'] = container_hash
    return metadata


def resolve_environment(task):
    """Use build interface to create a valid environment."""
    build_main(task)


@connect_db
def submit_jobs(db):
    """Submit all jobs currently in submit queue to the workload scheduler."""
    while db.submit_queue.count() >= 1:
        # Single value dictionary
        task = db.submit_queue.pop()
        try:
            log.info(f'Resolving environment for task {task.name}')
            resolve_environment(task)
            log.info(f'Environment preparation complete for task {task.name}')
            job_id, job_state = worker.submit_task(task)
            log.info(f'Job Submitted {task.name}: job_id: {job_id} job_state: {job_state}')
            # place job in queue to monitor
            db.job_queue.push(task=task, job_id=job_id, job_state=job_state)
            # update_task_metadata(task.id, task_metadata)
        except Exception as err:
            # Set job state to failed
            job_state = 'SUBMIT_FAIL'
            log.error(f'Task Manager submit task {task.name} failed! \n {err}')
            log.error(f'{task.name} state: {job_state}')
            # Log the traceback information as well
            log.error(traceback.format_exc())
        finally:
            # Send the initial state to WFM
            # update_task_state(task.id, job_state, metadata=task_metadata)
            update_task_state(task.workflow_id, task.id, job_state)


def get_task_checkpoint(task):
    """Harvest task checkpoint."""
    task_checkpoint = None
    hints = dict(task.hints)
    try:
        # Try to get Hints
        hint_checkpoint = hints['beeflow:CheckpointRequirement']
    except (KeyError, TypeError):
        # Task Hints are not mandatory. No task checkpoint hint specified.
        hint_checkpoint = None
    try:
        # Try to get Requirements
        req_checkpoint = task.requirements['beeflow:CheckpointRequirement']
    except (KeyError, TypeError):
        # Task Requirements are not mandatory. No task checkpoint requirement specified.
        req_checkpoint = None
    # Prefer requirements over hints
    if req_checkpoint:
        task_checkpoint = req_checkpoint
    elif hint_checkpoint:
        task_checkpoint = hint_checkpoint
    return task_checkpoint


def get_restart_file(task_checkpoint, task_workdir):
    """Find latest checkpoint file."""
    if 'file_regex' not in task_checkpoint:
        raise RuntimeError('file_regex is required for checkpointing')
    if 'file_path' not in task_checkpoint:
        raise RuntimeError('file_path is required for checkpointing')
    file_regex = task_checkpoint['file_regex']
    file_path = Path(task_workdir, task_checkpoint['file_path'])
    regex = re.compile(file_regex)
    checkpoint_files = [
        Path(file_path, fname) for fname in os.listdir(file_path)
        if regex.match(fname)
    ]
    checkpoint_files.sort(key=os.path.getmtime)
    try:
        checkpoint_file = checkpoint_files[-1]
        return str(checkpoint_file)
    except IndexError:
        raise RuntimeError('Missing checkpoint file for task') from None


@connect_db
def update_jobs(db):
    """Check and update states of jobs in queue, remove completed jobs."""
    # Need to make a copy first
    job_q = list(db.job_queue)
    for job in job_q:
        id_ = job['id']
        task = job['task']
        job_id = job['job_id']
        job_state = worker.query_task(job_id)

        # If state changes update the WFM
        if job_state != job['job_state']:
            log.info(f'{task.name} {job["job_state"]} -> {job_state}')
            # job['job_state'] = job_state
            db.job_queue.update_job_state(id_, job_state)
            if job_state in ('FAILED', 'TIMELIMIT', 'TIMEOUT'):
                # Harvest lastest checkpoint file.
                task_checkpoint = get_task_checkpoint(task)
                if task_checkpoint:
                    checkpoint_file = get_restart_file(task_checkpoint, task.workdir)
                    task_info = {'checkpoint_file': checkpoint_file, 'restart': True}
                    log.info(f'Restart: {task.name} task_info: {task_info}')
                    update_task_state(task.workflow_id, task.id, job_state, task_info=task_info)
                else:
                    update_task_state(task.workflow_id, task.id, job_state)
            else:
                update_task_state(task.workflow_id, task.id, job_state)

        if job_state in ('ZOMBIE', 'COMPLETED', 'CANCELLED', 'FAILED', 'TIMEOUT', 'TIMELIMIT'):
            # Remove from the job queue. Our job is finished
            db.job_queue.remove_by_id(id_)
            # job_queue.remove(job)
            log.info(f'Job {job_id} done {task.name}: removed from job status queue')


def process_queues():
    """Look for newly submitted jobs and update status of scheduled jobs."""
    submit_jobs()  # noqa
    update_jobs()  # noqa


if "pytest" not in sys.modules:
    scheduler = BackgroundScheduler({'apscheduler.timezone': 'UTC'})
    scheduler.add_job(func=process_queues, trigger="interval", seconds=8)
    scheduler.start()

    # This kills the scheduler when the process terminates
    # so we don't accidentally leave a zombie process
    atexit.register(scheduler.shutdown)


class TaskSubmit(Resource):
    """WFM sends tasks to the task manager."""

    def __init__(self):
        """Intialize request."""

    @staticmethod
    @connect_db
    def post(db):
        """Receives task from WFM."""
        parser = reqparse.RequestParser()
        parser.add_argument('tasks', type=str, location='json')
        data = parser.parse_args()
        tasks = jsonpickle.decode(data['tasks'])
        for task in tasks:
            db.submit_queue.push(task)
            log.info(f"Added {task.name} task to the submit queue")
        resp = make_response(jsonify(msg='Tasks Added!', status='ok'), 200)
        return resp


class TaskActions(Resource):
    """Actions to take for tasks."""

    @staticmethod
    @connect_db
    def delete(db):
        """Cancel received from WFM to cancel job, update queue to monitor state."""
        cancel_msg = ""
        for job in db.job_queue:
            task_id = job['task'].id
            job_id = job['job_id']
            name = job['task'].name
            log.info(f"Cancelling {name} with job_id: {job_id}")
            try:
                job_state = worker.cancel_task(job_id)
            except Exception as err:
                log.error(err)
                log.error(traceback.format_exc())
                job_state = 'ZOMBIE'
            cancel_msg += f"{name} {task_id} {job_id} {job_state}"
        db.job_queue.clear()
        db.submit_queue.clear()
        resp = make_response(jsonify(msg=cancel_msg, status='ok'), 200)
        return resp


# This could probably be in a Resource class, but since its only one route
# it seems to be fine right here
@flask_app.route('/status')
def get_status():
    """Report the current status of the Task Manager."""
    return make_response(jsonify(status='up'), 200)


WLS = bc.get('DEFAULT', 'workload_scheduler')
worker_class = worker_pkg.find_worker(WLS)
if worker_class is None:
    sys.exit(f'Workload scheduler {WLS}, not supported.\n'
             + f'Please check {bc.userconfig_path()} and restart TaskManager.')
# Get the parameters for the worker classes
worker_kwargs = {
    'bee_workdir': bc.get('DEFAULT', 'bee_workdir'),
    'container_runtime': bc.get('task_manager', 'container_runtime'),
    'job_template': bc.get('task_manager', 'job_template'),
    # extra options to be passed to the runner (i.e. srun [RUNNER_OPTS] ... for Slurm)
    'runner_opts': bc.get('task_manager', 'runner_opts'),
}
if WLS == 'Slurm':
    worker_kwargs['slurm_socket'] = bc.get('slurmrestd', 'slurm_socket')
worker = WorkerInterface(worker_class, **worker_kwargs)

api.add_resource(TaskSubmit, '/bee_tm/v1/task/submit/')
api.add_resource(TaskActions, '/bee_tm/v1/task/')

# if __name__ == '__main__':
#    hostname = socket.gethostname()
#    log.info(f'Starting Task Manager on host: {hostname}')
#    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
#    handler = bee_logging.save_log(bee_workdir=bee_workdir, log=log, logfile='task_manager.log')
#    log.info(f'tm_listen_port:{tm_listen_port}')
#    container_runtime = bc.get('task_manager', 'container_runtime')
#    log.info(f'container_runtime: {container_runtime}')
#
#    # Werkzeug logging
#    werk_log = logging.getLogger('werkzeug')
#    werk_log.setLevel(logging.INFO)
#    werk_log.addHandler(handler)
#
#    # Flask logging
#    flask_app.logger.addHandler(handler)
#    flask_app.run(debug=False, port=str(tm_listen_port))

# Ignoring CO413 beeflow modules must be loaded after bc.init()
# Ignoring W0703: Catching general exception is ok for failed submit and cancel.
# pylama:ignore=C0413,W0703



"""Tests of the TM database."""
import tempfile
import os

import pytest

from beeflow.common.db import tm


@pytest.fixture
def temp_db():
    """Pytest fixture for creating a temporary database."""
    fname = tempfile.mktemp()
    with tm.open_db(fname) as db:
        yield db
    os.remove(fname)


def test_empty(temp_db):
    """Test an empty database."""
    db = temp_db

    assert db.submit_queue.count() == 0
    assert db.job_queue.count() == 0


def test_push_pop(temp_db):
    """Test pushing and popping values in a database."""
    db = temp_db

    # Just push something random for the submit_queue, no need to be a real
    # task
    db.submit_queue.push(3)
    db.job_queue.push(task=45, job_id=1289, job_state='READY')

    assert db.submit_queue.count() == 1
    assert db.job_queue.count() == 1
    assert db.submit_queue.pop() == 3
    assert db.job_queue.pop() == {
        'task': 45,
        'job_id': 1289,
        'job_state': 'READY',
    }
    assert db.submit_queue.count() == 0
    assert db.job_queue.count() == 0


def test_push_pop_many(temp_db):
    """Test pushing and popping many values."""
    db = temp_db

    # Push 128 items
    for i in range(128):
        db.submit_queue.push(i)
        job_state = 'READY' if (i % 2) == 0 else 'COMPLETED'
        db.job_queue.push(task=i, job_id=i + 1, job_state=job_state)
        assert db.submit_queue.count() == (i + 1)
        assert db.job_queue.count() == (i + 1)
    # Now pop 128 items
    for i in range(128):
        assert db.submit_queue.pop() == i
        assert db.job_queue.pop() == {
            'task': i,
            'job_id': i + 1,
            'job_state': 'READY' if (i % 2) == 0 else 'COMPLETED',
        }
        assert db.submit_queue.count() == (127 - i)
        assert db.job_queue.count() == (127 - i)
    assert db.submit_queue.count() == 0
    assert db.job_queue.count() == 0


def test_clear(temp_db):
    """Test clearing the database."""
    db = temp_db

    # submit_queue
    db.submit_queue.push(3)
    db.submit_queue.push(4)
    db.submit_queue.push({123, 4, 5})
    db.submit_queue.clear()
    assert db.submit_queue.count() == 0

    # job_queue
    db.job_queue.push(task={1, 2, 3}, job_id=168, job_state='COMPLETED')
    db.job_queue.push(task={'some': 'value'}, job_id=12, job_state='READY')
    db.job_queue.push(task=(1, 2, 3), job_id=88888, job_state='FAILED')
    db.job_queue.clear()
    assert db.submit_queue.count() == 0


def test_iter(temp_db):
    """Test iterating over the values."""
    db = temp_db

    values = (127, 16, 11999)
    for val in values:
        db.submit_queue.push(val)
    for task, other in zip(db.submit_queue, values):
        assert task == other

    for val in values:
        db.job_queue.push(task=val, job_id=val, job_state='COMPLETED')
    for job, val in zip(db.job_queue, values):
        assert 'id' in job
        assert job['task'] == val
        assert job['job_id'] == val
        assert job['job_state'] == 'COMPLETED'


def test_job_queue_remove_by_id(temp_db):
    """Test removing a job by ID for the job queue."""
    db = temp_db

    db.job_queue.push(task={8, 9, 10}, job_id=888, job_state='some-state0')
    db.job_queue.push(task={10}, job_id=999, job_state='some-state1')
    db.job_queue.push(task={46, 57}, job_id=111, job_state='some-state2')

    assert db.job_queue.count() == 3

    jobs = list(db.job_queue)
    id_ = jobs[1]['id']
    db.job_queue.remove_by_id(id_)

    assert db.job_queue.count() == 2
    job = db.job_queue.pop()
    assert job['task'] == {8, 9, 10}
    assert job['job_id'] == 888
    assert job['job_state'] == 'some-state0'
    job = db.job_queue.pop()
    assert job['task'] == {46, 57}
    assert job['job_id'] == 111
    assert job['job_state'] == 'some-state2'
    assert db.job_queue.count() == 0


def test_job_queue_update_job_state(temp_db):
    """Test updating the job state for a job in the queue."""
    db = temp_db

    db.job_queue.push(task={8, 9, 10}, job_id='888', job_state='READY')

    jobs = list(db.job_queue)
    id_ = jobs[0]['id']
    db.job_queue.update_job_state(id_, 'COMPLETED')

    job = db.job_queue.pop()
    assert job['task'] == {8, 9, 10}
    assert job['job_id'] == 888
    assert job['job_state'] == 'COMPLETED'
# Ignore W0621: PyLama complains about redefining 'temp_db' from the outer
#               scope. This is how pytest fixtures work.
# pylama:ignore=W0621



#!/usr/bin/env python3

"""beeflow.

This script manages the startup of the BEE daemons and supporting services.
If no arguments are given this script will start the BEEWorkflowManager,
BEETaskManager, and all required supporting services. If any combination of
services is specified using the appropriate flag(s) then ONLY those services
will be started.
"""
import os
import signal
import subprocess
import socket
import sys
import shutil
import time
import importlib.metadata

import daemon
import typer

from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common import cli_connection


class ComponentManager:
    """Component manager class."""

    def __init__(self):
        """Construct the component manager."""
        self.components = {}
        self.procs = {}

    def component(self, name, deps=None):
        """Return a decorator function to be called."""

        def wrap(fn):
            """Add the component to the list."""
            self.components[name] = {
                'fn': fn,
                'deps': deps,
            }

        return wrap

    def _validate(self, base_components):
        """Make sure that the components all exist and have valid deps."""
        missing = [name for name in base_components if name not in self.components]
        for component in self.components.values():
            if component['deps'] is None:
                continue
            for dep in component['deps']:
                if dep not in self.components:
                    missing.append(dep)
        if missing:
            raise RuntimeError(f'Missing/unknown component(s): {",".join(missing)}')

    def _find_order(self, base_components):
        """Find the order of the dependencies to launch."""
        s = base_components[:]
        levels = {name: 0 for name in self.components}
        while len(s) > 0:
            name = s.pop()
            if self.components[name]['deps'] is None:
                continue
            for dep in self.components[name]['deps']:
                levels[dep] = max(levels[name] + 1, levels[dep])
                # Detect a possible cycle
                if levels[dep] > len(self.components):
                    raise RuntimeError(f'There may be a cycle for the "{dep}" component')
                s.append(dep)
        levels = list(levels.items())
        levels.sort(key=lambda t: t[1], reverse=True)
        return [name for name, level in levels]

    def run(self, base_components):
        """Start and run everything."""
        # Determine if there are any missing components listed
        self._validate(base_components)
        # Determine the order to launch components in (note: this should just ignore cycles)
        order = self._find_order(base_components)
        print(f'Launching components in order: {order}')
        # Now launch the components
        for name in order:
            component = self.components[name]
            self.procs[name] = component['fn']()

    def poll(self):
        """Poll each process to check for errors."""
        for name, proc in self.procs.items():
            returncode = proc.poll()
            if returncode is not None:
                log = log_fname(name)
                print(f'Component "{name}" failed, check log "{log}"')

    def status(self):
        """Return the statuses for each process in a dict."""
        return {
            name: 'RUNNING' if proc.poll() is None else 'FAILED'
            for name, proc in self.procs.items()
        }

    def kill(self):
        """Kill all components."""
        for name, proc in self.procs.items():
            print(f'Killing {name}')
            proc.terminate()


MGR = ComponentManager()


def warn(*pargs):
    """Print a red warning message."""
    typer.secho(' '.join(pargs), fg=typer.colors.RED, file=sys.stderr)


def launch_with_gunicorn(module, sock_path, *args, **kwargs):
    """Launch a component with Gunicorn."""
    # Setting the timeout to infinite, since sometimes the gdb can take too long
    return subprocess.Popen(['gunicorn', module, '--timeout', '0', '-b', f'unix:{sock_path}'],
                            *args, **kwargs)


def log_path():
    """Return the main log path."""
    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    return os.path.join(bee_workdir, 'logs')


def log_fname(component):
    """Determine the log file name for the given component."""
    return os.path.join(log_path(), f'{component}.log')


def open_log(component):
    """Determine the log for the component, open and return it."""
    log = log_fname(component)
    return open(log, 'a', encoding='utf-8')


@MGR.component('wf_manager', ('scheduler',))
def start_wfm():
    """Start the WFM."""
    fp = open_log('wf_manager')
    sock_path = bc.get('workflow_manager', 'socket')
    return launch_with_gunicorn('beeflow.wf_manager.wf_manager:create_app()',
                                sock_path, stdout=fp, stderr=fp)


@MGR.component('task_manager', ('slurmrestd',))
def start_task_manager():
    """Start the TM."""
    fp = open_log('task_manager')
    sock_path = bc.get('task_manager', 'socket')
    return launch_with_gunicorn('beeflow.task_manager:flask_app', sock_path, stdout=fp, stderr=fp)


@MGR.component('scheduler', ())
def start_scheduler():
    """Start the scheduler."""
    fp = open_log('scheduler')
    sock_path = bc.get('scheduler', 'socket')
    # Using a function here because of the funny way that the scheduler's written
    return launch_with_gunicorn('beeflow.scheduler.scheduler:create_app()', sock_path, stdout=fp,
                                stderr=fp)


# Workflow manager and task manager need to be opened with PIPE for their stdout/stderr
@MGR.component('slurmrestd')
def start_slurm_restd():
    """Start BEESlurmRestD. Returns a Popen process object."""
    bee_workdir = bc.get('DEFAULT', 'bee_workdir')
    slurmrestd_log = '/'.join([bee_workdir, 'logs', 'restd.log'])
    slurm_socket = bc.get('slurmrestd', 'slurm_socket')
    slurm_args = bc.get('slurmrestd', 'slurm_args')
    slurm_args = slurm_args if slurm_args is not None else ''
    subprocess.run(['rm', '-f', slurm_socket], check=True)
    # log.info("Attempting to open socket: {}".format(slurm_socket))
    fp = open(slurmrestd_log, 'w', encoding='utf-8') # noqa
    cmd = ['slurmrestd']
    cmd.extend(slurm_args.split())
    cmd.append(f'unix:{slurm_socket}')
    return subprocess.Popen(cmd, stdout=fp, stderr=fp)


def handle_terminate(signum, stack): # noqa
    """Handle a terminate signal."""
    # Kill all subprocesses
    MGR.kill()
    sys.exit(1)


MIN_CHARLIECLOUD_VERSION = (0, 27)


def version_str(version):
    """Convert a version tuple to a string."""
    return '.'.join([str(part) for part in version])


def check_dependencies():
    """Check for various dependencies in the environment."""
    print('Checking dependencies...')
    # Check for Charliecloud and it's version
    if not shutil.which('ch-run'):
        warn('Charliecloud is not loaded. Please ensure that it is accessible on your path.')
        sys.exit(1)
    cproc = subprocess.run(['ch-run', '-V'], capture_output=True, text=True,
                           check=True)
    version = cproc.stdout if cproc.stdout else cproc.stderr
    version = version.strip()
    version = tuple(int(part) for part in version.split('.'))
    print(f'Found Charliecloud {version_str(version)}')
    if version < MIN_CHARLIECLOUD_VERSION:
        warn('This version of Charliecloud is too old, please upgrade to at '
             f'least version {version_str(MIN_CHARLIECLOUD_VERSION)}')
        sys.exit(1)


class Beeflow:
    """Beeflow class for handling the main loop."""

    def __init__(self, mgr, base_components):
        """Create the Beeflow class."""
        self.mgr = mgr
        self.base_components = base_components
        self.quit = False

    def loop(self):
        """Run the main loop."""
        print(f'Running on {socket.gethostname()}')
        self.mgr.run(self.base_components)
        sock_path = bc.get('DEFAULT', 'beeflow_socket')
        with cli_connection.server(sock_path) as server:
            while not self.quit:
                # Handle a message from the client, if there is one
                self.handle_client(server)
                # Poll the components
                self.mgr.poll()
                time.sleep(1)
        # Kill everything, if possible
        self.mgr.kill()

    def handle_client(self, server):
        """Handle a message from the client."""
        try:
            client = server.accept()
            if client is None:
                return
            msg = client.get()
            resp = None
            if msg['type'] == 'status':
                resp = {
                    'components': self.mgr.status(),
                }
                print('Returned status info.')
            elif msg['type'] == 'quit':
                self.quit = True
                resp = 'shutting down'
                print('Shutting down.')
            client.put(resp)
        except cli_connection.BeeflowConnectionError as err:
            print(f'connection failed: {err}')


def daemonize(base_components):
    """Start beeflow as a daemon, monitoring all processes."""
    # Now set signal handling, the log and finally daemonize
    signal_map = {
        signal.SIGINT: handle_terminate,
        signal.SIGTERM: handle_terminate,
    }
    fp = open_log('beeflow')
    with daemon.DaemonContext(signal_map=signal_map, stdout=fp, stderr=fp, stdin=fp,
                              umask=0o002):
        Beeflow(MGR, base_components).loop()


app = typer.Typer(no_args_is_help=True)


@app.command()
def start(foreground: bool = typer.Option(False, '--foreground', '-F',
          help='run in the foreground')):
    """Attempt to daemonize if not in debug and start all BEE components."""
    beeflow_log = log_fname('beeflow')
    check_dependencies()
    sock_path = bc.get('DEFAULT', 'beeflow_socket')
    # Note: there is a possible race condition here, however unlikely
    if os.path.exists(sock_path):
        # Try to contact for a status
        resp = cli_connection.send(sock_path, {'type': 'status'})
        if resp is None:
            # Must be dead, so remove the socket path
            os.remove(sock_path)
        else:
            # It's already running, so print an error and exit
            warn(f'Beeflow appears to be running. Check the beeflow log: "{beeflow_log}"')
            sys.exit(1)
    print('Starting beeflow...')
    if not foreground:
        print(f'Check "{beeflow_log}" or run `beeflow status` for more information.')
    # Create the log path if it doesn't exist yet
    path = log_path()
    os.makedirs(path, exist_ok=True)
    base_components = ['wf_manager', 'task_manager', 'scheduler']
    if foreground:
        try:
            Beeflow(MGR, base_components).loop()
        except KeyboardInterrupt:
            MGR.kill()
    else:
        daemonize(base_components)


@app.command()
def status():
    """Check the status of beeflow and the components."""
    sock_path = bc.get('DEFAULT', 'beeflow_socket')
    resp = cli_connection.send(sock_path, {'type': 'status'})
    if resp is None:
        beeflow_log = log_fname('beeflow')
        warn('Cannot connect to the beeflow daemon, is it running? Check the '
             f'log at "{beeflow_log}".')
        sys.exit(1)
    print('beeflow components:')
    for comp, stat in resp['components'].items():
        print(f'{comp} ... {stat}')


@app.command()
def stop():
    """Stop the current running beeflow daemon."""
    stop_msg = ("\n** Please ensure all workflows are complete before stopping beeflow. **"
                + "\n** Check the status of workflows by running 'beeclient listall'.    **"
                + "\nAre you sure you want to kill beeflow components? [y/n] ")
    ans = input(stop_msg)
    if ans.lower() != 'y':
        return
    sock_path = bc.get('DEFAULT', 'beeflow_socket')
    resp = cli_connection.send(sock_path, {'type': 'quit'})
    if resp is None:
        beeflow_log = log_fname('beeflow')
        warn('Error: beeflow is not running on this system. It could be '
             'running on a different front end.\n'
             f'       Check the beeflow log: "{beeflow_log}".')
        sys.exit(1)
    # As long as it returned something, we should be good
    beeflow_log = log_fname('beeflow')
    print(f'Beeflow has stopped. Check the log at "{beeflow_log}".')


@app.callback(invoke_without_command=True)
def version_callback(version: bool = False):
    """Beeflow."""
    # Print out the current version of the app, and then exit
    # Note above docstring gets used in the help menu
    if version:
        version = importlib.metadata.version("hpc-beeflow")
        print(version)


def main():
    """Start the beeflow app."""
    bc.init()
    app()


if __name__ == '__main__':
    main()



"""Abstract base class for the handling of workflow DAGs."""

from abc import ABC, abstractmethod


class GraphDatabaseDriver(ABC):
    """Driver interface for a generic graph database.

    The driver must implement a __init__ method that creates/connects to
    the graph database and returns some kind of 'connection' interface object.
    """

    @abstractmethod
    def initialize_workflow(self, workflow):
        """Begin construction of a workflow in the graph database.

        Should create the Workflow, Requirement, and Hint nodes in the graph database.

        :param workflow: the workflow description
        :type workflow: Workflow
        """

    @abstractmethod
    def execute_workflow(self):
        """Begin execution of the stored workflow.

        Set the initial tasks' states to 'READY'.
        """

    @abstractmethod
    def pause_workflow(self):
        """Pause execution of a running workflow.

        Set workflow from state 'RUNNING' to 'PAUSED'.
        """

    @abstractmethod
    def resume_workflow(self):
        """Resume execution of a paused workflow.

        Set workflow state from 'PAUSED' to 'RUNNING'.
        """

    @abstractmethod
    def reset_workflow(self, new_id):
        """Reset the execution state of a stored workflow.

        Set all task states to 'WAITING'.
        Change the workflow ID of the Workflow and Task nodes to new_id.
        Delete all task metadata except for task state.

        :param new_id: the new workflow ID
        :type new_id: str
        """

    @abstractmethod
    def load_task(self, task):
        """Load a task into a stored workflow.

        Dependencies should be automatically deduced and generated by the graph database
        upon loading each task by matching workflow inputs with new task inputs,
        or task outputs with new task inputs.

        :param task: a workflow task
        :type task: Task
        """

    @abstractmethod
    def initialize_ready_tasks(self):
        """Set runnable tasks to state 'READY'.

        Runnable tasks are tasks with all input dependencies fulfilled.
        """

    @abstractmethod
    def restart_task(self, old_task, new_task):
        """Restart a failed task.

        Create a Task node for new_task with state 'RESTARTED' and an edge
        to indicate that it is the child of the Task node of old_task.

        :param old_task: the failed task
        :type old_task: Task
        :param new_task: the new (restarted) task
        :type new_task: Task
        """

    @abstractmethod
    def finalize_task(self, task):
        """Set task state to 'COMPLETED' and set inputs from source.

        :param task: the task to finalize
        :type task: Task
        """

    @abstractmethod
    def get_task_by_id(self, task_id):
        """Return a reconstructed Task object from the graph database by its ID.

        :param task_id: a task's ID
        :type task_id: str
        :rtype: Task
        """

    @abstractmethod
    def get_workflow_description(self):
        """Return a reconstructed Workflow object from the graph database.

        :rtype: Workflow
        """

    @abstractmethod
    def get_workflow_state(self):
        """Return the current state of the workflow.

        :rtype: str
        """

    @abstractmethod
    def set_workflow_state(self, state):
        """Set the state of the workflow.

        :param state: the new state of the workflow
        :type state: str
        """

    @abstractmethod
    def get_workflow_tasks(self):
        """Return a list of all workflow tasks from the graph database.

        :rtype: list of Task
        """

    @abstractmethod
    def get_workflow_requirements_and_hints(self):
        """Return all workflow requirements and hints from the graph database.

        Must return a tuple with the format (requirements, hints)

        :rtype: (list of Requirement, list of Hint)
        """

    @abstractmethod
    def get_workflow_inputs_and_outputs(self):
        """Return all workflow inputs and outputs from the graph database.

        Returns a tuple of (inputs, outputs).

        :rtype: (list of InputParameter, list of OutputParameter)
        """

    @abstractmethod
    def get_ready_tasks(self):
        """Return tasks with state 'READY' from the graph database.

        :rtype: list of Task
        """

    @abstractmethod
    def get_dependent_tasks(self, task):
        """Return the dependent tasks of a workflow task in the graph database.

        :param task: the task whose dependents to retrieve
        :type task: Task
        :rtype: list of Task
        """

    @abstractmethod
    def get_task_state(self, task):
        """Return the state of a task in the graph database.

        :param task: the task whose status to retrieve
        :type task: Task
        :rtype: str
        """

    @abstractmethod
    def set_task_state(self, task, state):
        """Set the state of a task in the graph database.

        :param task: the task whose state to set
        :type task: Task
        :param state: the new state
        :type state: str
        """

    @abstractmethod
    def get_task_metadata(self, task):
        """Return the metadata of a task in the graph database.

        :param task: the task whose metadata to retrieve
        :type task: Task
        :rtype: dict
        """

    @abstractmethod
    def set_task_metadata(self, task, metadata):
        """Set the metadata of a task in the graph database.

        :param task: the task whose metadata to set
        :type task: Task
        :param metadata: the job description metadata
        :type metadata: dict
        """

    @abstractmethod
    def get_task_input(self, task, input_id):
        """Get a task input object.

        :param task: the task whose input to retrieve
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :rtype: StepInput
        """

    @abstractmethod
    def set_task_input(self, task, input_id, value):
        """Set the value of a task input.

        :param task: the task whose input to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param value: str or int or float
        """

    @abstractmethod
    def get_task_output(self, task, output_id):
        """Get a task output object.

        :param task: the task whose output to retrieve
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :rtype: StepOutput
        """

    @abstractmethod
    def set_task_output(self, task, output_id, value):
        """Set the value of a task output.

        :param task: the task whose output to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param value: the output value to set
        :type value: str or int or float
        """

    @abstractmethod
    def set_task_input_type(self, task, input_id, type_):
        """Set the type of a task input.

        :param task: the task whose input type to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param type_: the input type to set
        :param type_: str
        """

    @abstractmethod
    def set_task_output_glob(self, task, output_id, glob):
        """Set the glob of a task output.

        :param task: the task whose output glob to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param glob: the output glob to set
        :type glob: str
        """

    @abstractmethod
    def workflow_completed(self):
        """Determine if a workflow has completed.

        A workflow has completed if each of its final tasks has state 'COMPLETED'.

        :rtype: bool
        """

    @abstractmethod
    def empty(self):
        """Determine if the database is empty.

        :rtype: bool
        """

    @abstractmethod
    def cleanup(self):
        """Clean up all the data stored in the graph database."""

    @abstractmethod
    def close(self):
        """Close the connection to the graph database."""



"""Cloud init module."""
from beeflow.common.cloud import chameleoncloud
from beeflow.common.cloud import openstack
from beeflow.common.cloud import provider
from beeflow.common.cloud import google
from beeflow.common.cloud.constants import BEE_USER
from beeflow.common.cloud.cloud import CloudError


providers = {
    'google': google.GoogleProvider,
    'chameleoncloud': chameleoncloud.ChameleoncloudProvider,
    'openstack': openstack.OpenstackProvider,
    'mock': provider.MockProvider,  # Provider to be used for testing
}


def get_provider(name, **kwargs):
    """Return a Provider object for the given provider."""
    if name in providers:
        return providers[name](**kwargs)

    raise RuntimeError(f'Invalid provider "{name}"')
# Ignore W0611: These are meant to be used by external code
# pylama:ignore=W0611



"""Abstract base class for the handling build systems."""

from abc import ABC, abstractmethod
import jsonpickle


def arg2task(task_arg):
    """Convert JSON encoded task to Task object.

    The build driver will expect a Task object, and the build
    interface starts with a JSON representation of the Task object.
    """
    return jsonpickle.decode(task_arg)


def task2arg(task):
    """Convert Task object to JSON encoded string.

    The build interface needs to pass Task data on the command line,
    because each compute node needs to understand the Task description.
    JSON format is a convenient way to describe the Task object at the
    command line.
    """
    return jsonpickle.encode(task)


class BuildDriver(ABC):
    """Driver interface between WFM and a generic build system.

    A driver object must implement an __init__ method that
    requests a Runtime Environment (RTE), and a method to
    return the requested RTE.
    """

    @abstractmethod
    def __init__(self, task, kwargs):
        """Begin build request.

        Parse hints and requirements to determine target build
        system. Harvest relevant BeeConfig params in kwargs.

        :param requirements: the workflow requirements
        :type requirements: set of Requirement instances
        :param hints: the workflow hints (optional requirements)
        :type hints: set of Requirement instances
        :param kwargs: Dictionary of build system config options
        :type kwargs: set of build system parameters
        """

    @abstractmethod
    def process_docker_pull(self, addr, force):
        """Get and process the CWL compliant dockerPull dockerRequirement.

        CWL spec 09-23-2020: Specify a Docker image to
        retrieve using docker pull. Can contain the immutable
        digest to ensure an exact container is used.
        """

    @abstractmethod
    def process_docker_load(self):
        """Get and process the CWL compliant dockerLoad dockerRequirement.

        CWL spec 09-23-2020: Specify a HTTP URL from which to
        download a Docker image using docker load.
        """

    @abstractmethod
    def process_docker_file(self, task_dockerfile, force):
        """Get and process the CWL compliant dockerFile dockerRequirement.

        CWL spec 09-23-2020: Supply the contents of a Dockerfile
        which will be built using docker build.
        """

    @abstractmethod
    def process_docker_import(self, param_import):
        """Get and process the CWL compliant dockerImport dockerRequirement.

        CWL spec 09-23-2020: Provide HTTP URL to download and
        gunzip a Docker images using docker import.
        """

    @abstractmethod
    def process_docker_image_id(self, param_imageid):
        """Get and process the CWL compliant dockerImageId dockerRequirement.

        CWL spec 09-23-2020: The image id that will be used for
        docker run. May be a human-readable image name or the
        image identifier hash. May be skipped if dockerPull is
        specified, in which case the dockerPull image id must be
        used.
        """

    @abstractmethod
    def process_docker_output_directory(self, param_output_directory):
        """Get and process the CWL compliant dockerOutputDirectory dockerRequirement.

        CWL spec 09-23-2020: Set the designated output directory
        to a specific location inside the Docker container.
        """

    @abstractmethod
    def process_copy_container(self, force):
        """Get and process the BEE CWL extension copyContainer dockerRequirement.

        This CWL extension will copy an existing container to the build archive.

        If you have a container tarball, and all you need to do is stage it,
        that is, all you need to do is copy it to a location that BEE knows,
        use this to put the container into the build archive.
        """

    @abstractmethod
    def process_container_name(self):
        """Get and process BEE CWL extension for containerName dockerRequirement.

        This is a BEE extension to CWL to refer to containers with human-readable name.

        The CWL spec currently uses dockerImageId to refer to the name of a container
        but this is explicitly not how Docker defines it. We need a way to name
        containers in a human readable format.
        """

    def resolve_priority(self):
        """Given multiple DockerRequirements, set order of execution.

        The CWL spec as of 04-15-2021 does not specify order of
        execution, but the cwltool gives some guidance by example.
        We mimic cwltool in how we resolve priority, favoring
        fast, cached container specs over slower specs. For example,
        if both a docker pull and a docker file are supported, the
        build interface will try to pull first, and only on pull
        failure will the builder build the docker file.
        """
        # cwl spec priority list consists of:
        # (bound method, method name, priority, terminal case bool)
        cwl_spec = [(self.process_docker_pull, 'dockerPull', 5, True),
                    (self.process_docker_load, 'dockerLoad', 6, True),
                    (self.process_docker_file, 'dockerFile', 7, True),
                    (self.process_docker_import, 'dockerImport', 4, True),
                    (self.process_copy_container, 'beeflow:copyContainer', 3, True),
                    (self.process_docker_image_id, 'dockerImageId', 1, False),
                    (self.process_container_name, 'beeflow:containerName', 2, False),
                    (self.process_docker_output_directory, 'dockerOutputDirectory', 0, False)]
        exec_list = sorted(cwl_spec, key=lambda x: x[2])
        return exec_list



"""Common CLI code used by different BEE scripts."""
import click


class NaturalOrderGroup(click.Group):
    """Natural ordering class for using with CLI code."""

    def list_commands(self, ctx):  # noqa
        """List the commands in order."""
        return self.commands.keys()



"""Init file for the worker package."""

from beeflow.common.worker.slurm_worker import SlurmWorker
from beeflow.common.worker.lsf_worker import LSFWorker
from beeflow.common.worker.simple_worker import SimpleWorker


supported_workload_schedulers = {
    'Slurm': SlurmWorker,
    'LSF': LSFWorker,
    'Simple': SimpleWorker,
}


def find_worker(name):
    """Find the worker class or return None."""
    if name in supported_workload_schedulers:
        return supported_workload_schedulers[name]
    return None



"""Tests of the Slurm worker."""
import uuid
import time
import subprocess
import os
import pytest
from beeflow.common.config_driver import BeeConfig as bc


bc.init()


from beeflow.common.worker_interface import WorkerInterface
from beeflow.common.worker.worker import WorkerError
from beeflow.common.worker.slurm_worker import SlurmWorker
from beeflow.common.wf_data import Task


# Timeout (seconds) for waiting on tasks
TIMEOUT = 150
# Extra slurmrestd arguments. This may be something to take on the command line
SLURMRESTD_ARGS = bc.get('slurmrestd', 'slurm_args')
GOOD_TASK = Task(name='good-task', base_command=['sleep', '3'], hints=[],
                 requirements=[], inputs=[], outputs=[], stdout='', stderr='',
                 workflow_id=uuid.uuid4().hex)
BAD_TASK = Task(name='bad-task', base_command=['/this/is/not/a/command'], hints=[],
                requirements=[], inputs=[], outputs=[], stdout='', stderr='',
                workflow_id=uuid.uuid4().hex)


def wait_state(worker_iface, job_id, state):
    """Wait for Slurm to switch the job to another state."""
    time.sleep(1)
    n = 1
    last_state = worker_iface.query_task(job_id)
    while last_state == state:
        if n >= TIMEOUT:
            raise RuntimeError(f'job timed out, still in state {state}')
        time.sleep(1)
        n += 1
        last_state = worker_iface.query_task(job_id)
    return last_state


def setup_slurm_worker(fn):
    """Add a decorator to set up the worker interface and slurmrestd."""

    def decorator():
        """Decorate the input function."""
        slurm_socket = f'/tmp/{uuid.uuid4().hex}.sock'
        bee_workdir = os.path.expanduser(f'~/{uuid.uuid4().hex}.tmp')
        os.mkdir(bee_workdir)
        proc = subprocess.Popen(f'slurmrestd {SLURMRESTD_ARGS} unix:{slurm_socket}', shell=True)
        time.sleep(1)
        worker_iface = WorkerInterface(worker=SlurmWorker, container_runtime='Charliecloud',
                                       slurm_socket=slurm_socket, bee_workdir=bee_workdir,
                                       job_template=bc.get('task_manager', 'job_template'))
        fn(worker_iface)
        time.sleep(1)
        proc.kill()

    return decorator


def setup_worker_iface(fn):
    """Add a decorator that creates the worker interface but not slurmrestd."""

    def decorator():
        """Decorate the input function."""
        slurm_socket = f'/tmp/{uuid.uuid4().hex}.sock'
        bee_workdir = os.path.expanduser(f'~/{uuid.uuid4().hex}.tmp')
        os.mkdir(bee_workdir)
        worker_iface = WorkerInterface(worker=SlurmWorker, container_runtime='Charliecloud',
                                       slurm_socket=slurm_socket, bee_workdir=bee_workdir,
                                       job_template=bc.get('task_manager', 'job_template'))
        fn(worker_iface)

    return decorator


@setup_slurm_worker
def test_good_task(worker_iface):
    """Test submission of a good task."""
    job_id, last_state = worker_iface.submit_task(GOOD_TASK)
    assert last_state == 'PENDING'
    last_state = wait_state(worker_iface, job_id, 'PENDING')
    if last_state == 'RUNNING':
        last_state = wait_state(worker_iface, job_id, 'RUNNING')
    if last_state == 'COMPLETING':
        last_state = wait_state(worker_iface, job_id, 'COMPLETING')
    assert last_state == 'COMPLETED'


@setup_slurm_worker
def test_bad_task(worker_iface):
    """Test submission of a bad task."""
    job_id, last_state = worker_iface.submit_task(BAD_TASK)
    assert last_state == 'PENDING'
    last_state = wait_state(worker_iface, job_id, 'PENDING')
    if last_state == 'RUNNING':
        last_state = wait_state(worker_iface, job_id, 'RUNNING')
    if last_state == 'COMPLETING':
        last_state = wait_state(worker_iface, job_id, 'COMPLETING')
    assert last_state == 'FAILED'


@setup_slurm_worker
def test_query_bad_job_id(worker_iface):
    """Test querying a bad job ID."""
    with pytest.raises(WorkerError):
        worker_iface.query_task(888)


@setup_slurm_worker
def test_cancel_good_job(worker_iface):
    """Cancel a good job."""
    job_id, _ = worker_iface.submit_task(GOOD_TASK)
    job_state = worker_iface.cancel_task(job_id)
    assert job_state == 'CANCELLED'


# This test is broken in CI, but works locally. I'm commenting it out for now
# @setup_slurm_worker
# def test_cancel_bad_job_id(worker_iface):
#    """Cancel a non-existent job."""
#    with pytest.raises(WorkerError):
#        worker_iface.cancel_task(888)


@setup_worker_iface
def test_no_slurmrestd(worker_iface):
    """Test without running slurmrestd."""
    job_id, state = worker_iface.submit_task(GOOD_TASK)
    assert state == 'NOT_RESPONDING'
    assert worker_iface.query_task(job_id) == 'NOT_RESPONDING'
    assert worker_iface.cancel_task(job_id) == 'NOT_RESPONDING'
# Ignoring R1732: This is not what we need to do with the Popen of slurmrestd above;
#                 using a with statement doesn't kill the process immediately but just
#                 waits for it to complete and slurmrestd never will unless we kill it.
# pylama:ignore=R1732



"""Defines data structures for holding task and workflow data."""
from collections import namedtuple
from uuid import uuid4
from copy import deepcopy
import os

from beeflow.common.container_path import convert_path

# Workflow input parameter class
InputParameter = namedtuple("InputParameter", ["id", "type", "value"])
# Workflow output parameter class
OutputParameter = namedtuple("OutputParameter", ["id", "type", "value", "source"])
# Step input class
StepInput = namedtuple("StepInput", ["id", "type", "value", "default", "source", "prefix",
                                     "position", "value_from"])
# Step output class
StepOutput = namedtuple("StepOutput", ["id", "type", "value", "glob"])

# CWL requirement class
Requirement = namedtuple("Requirement", ["class_", "params"])
# CWL hint class
Hint = namedtuple("Hint", ["class_", "params"])


def generate_workflow_id():
    """Generate a unique workflow ID.

    :rtype: str
    """
    return uuid4().hex


class Workflow:
    """Data structure for holding data about a workflow."""

    def __init__(self, name, hints, requirements, inputs, outputs, workflow_id):
        """Store a workflow description.

        :param name: the workflow name
        :type name: string
        :param hints: the workflow hints
        :type hints: list of Requirements
        :param requirements: the workflow requirements
        :type requirements: list of Requirements
        :param inputs: the workflow inputs
        :type inputs: list of InputParameter
        :param outputs: the workflow outputs
        :type outputs: list of OutputParameter
        :param workflow_id: the workflow ID
        :type workflow_id: str
        """
        self.name = name
        self.hints = hints
        self.requirements = requirements
        self.inputs = inputs
        self.outputs = outputs
        self.id = workflow_id
        self.state = "SUBMITTED"

    def __eq__(self, other):
        """Test the equality of two workflows.

        Workflow ID and dependencies do not factor into equality testing.
        Currently, the code is boilerplate. We do not support multiple workflows.

        :param other: the workflow with which to test equality
        :type other: Workflow
        """
        if not isinstance(other, Workflow):
            return False

        def id_sort(i):
            return i.id

        return bool(self.name == other.name
                    and sorted(self.hints) == sorted(other.hints)
                    and sorted(self.requirements) == sorted(other.requirements)
                    and sorted(self.inputs, key=id_sort) == sorted(other.inputs, key=id_sort)
                    and sorted(self.outputs, key=id_sort) == sorted(other.outputs, key=id_sort))

    def __ne__(self, other):
        """Test the inequality of two workflows.

        :param other: the workflow with which to test inequality
        :type other: Workflow
        """
        return bool(not self.__eq__(other))

    def __repr__(self):
        """Construct a workflow's string representation."""
        return (f"<Workflow id={self.id} name={self.name} hints={self.hints} "
                f"requirements = {self.requirements} inputs={self.inputs} outputs={self.outputs}>")


class Task:
    """Data structure for holding data about a single task."""

    def __init__(self, name, base_command, hints, requirements, inputs, outputs, stdout, stderr,
                 workflow_id, task_id=None, workdir=None):
        """Store a task description.

        Task ID should only be given as a parameter when reconstructing the Task object
        from the graph database.

        :param name: the task name
        :type name: str
        :param base_command: the base command to run for the task
        :type base_command: str or list of str
        :param hints: the task hints (optional requirements)
        :type hints: list of Hint
        :param requirements: the task requirements
        :type requirements: list of Requirement
        :param inputs: the task inputs
        :type inputs: list of StepInput
        :param outputs: the task outputs
        :type outputs: list of StepOutput
        :param stdout: the name of the file to which to redirect stdout
        :type stdout: str
        :param stderr: the name of the file to which to redirect stderr
        :type stderr: str
        :param workflow_id: the workflow ID
        :type workflow_id: str
        :param task_id: the task ID
        :type task_id: str
        :param workdir: the working directory from which to get and store data
        :type workdir: path
        """
        self.name = name
        self.base_command = base_command
        self.hints = hints
        self.requirements = requirements
        self.inputs = inputs
        self.outputs = outputs
        self.stdout = stdout
        self.stderr = stderr
        self.workflow_id = workflow_id
        self.workdir = workdir

        # Task ID as UUID if not given
        if task_id is None:
            self.id = self.generate_task_id()
        else:
            self.id = task_id

    def generate_task_id(self):
        """Generate a unique task ID.

        :rtype: str
        """
        return uuid4().hex

    def copy(self, new_id=False):
        """Make a copy of this task.

        :param new_id: generate a new task ID
        :type new_id: bool
        :rtype: Task
        """
        task_id = self.generate_task_id() if new_id else self.id
        return Task(name=self.name, base_command=self.base_command,
                    hints=deepcopy(self.hints), requirements=deepcopy(self.requirements),
                    inputs=deepcopy(self.inputs), outputs=deepcopy(self.outputs),
                    stdout=self.stdout, stderr=self.stderr, workflow_id=self.workflow_id,
                    task_id=task_id, workdir=self.workdir)

    def get_requirement(self, req_type, req_param):
        """Get requirement from hints or requirements, prioritizing requirements over hints.

        :param req_type: the type of requirement (e.g. 'DockerRequirement')
        :type req_type: str
        :param req_param: the requirement parameter (e.g. 'dockerFile')
        :type req_param: str

        When requirements are specified hints will be ignored.
        By default, tasks need not specify hints or requirements
        """
        requirements = dict(self.requirements)
        requirement = None
        # Get value if specified in requirements
        try:
            # Try to get Requirement
            requirement = requirements[req_type][req_param]
        except (KeyError, TypeError):
            # Task Requirements are not mandatory. No docker_req_param specified in task reqs.
            requirement = None
        # Ignore hints if requirements available
        if not requirement:
            hints = dict(self.hints)
            # Get value if specified in hints
            try:
                # Try to get Hints
                requirement = hints[req_type][req_param]
            except (KeyError, TypeError):
                # Task Hints are not mandatory. No docker_req_param specified in task hints.
                requirement = None
        return requirement

    def __eq__(self, other):
        """Test the equality of two tasks.

        Task ID and dependencies do not factor into equality testing.
        :param other: the task with which to test equality
        :type other: Task
        :rtype: bool
        """
        if not isinstance(other, Task):
            return False

        def id_sort(i):
            return i.id

        return bool(self.name == other.name
                    and self.base_command == other.base_command
                    and sorted(self.hints) == sorted(other.hints)
                    and sorted(self.requirements) == sorted(other.requirements)
                    and sorted(self.inputs, key=id_sort) == sorted(other.inputs, key=id_sort)
                    and sorted(self.outputs, key=id_sort) == sorted(other.outputs, key=id_sort)
                    and self.stdout == other.stdout
                    and self.stderr == other.stderr
                    and self.workdir == other.workdir)

    def __ne__(self, other):
        """Test the inequality of two tasks.

        :param other: the task with which to test inequality
        :type other: Task
        :rtype: bool
        """
        return bool(not self.__eq__(other))

    def __hash__(self):
        """Return the hash value for a task.

        :rtype: int
        """
        return hash((self.id, self.workflow_id, self.name))

    def __repr__(self):
        """Construct a task's string representation.

        :rtype: str
        """
        return (f"<Task id={self.id} name={self.name} base_command={self.base_command} "
                f"hints={self.hints} requirements = {self.requirements} "
                f"inputs={self.inputs} outputs={self.outputs} stdout={self.stdout} "
                f"stderr={self.stderr} workflow_id={self.workflow_id}> ")

    @property
    def command(self):
        """Construct a task's command as a list.

        :rtype: list of str
        """
        positional_inputs = []
        nonpositional_inputs = []
        for input_ in self.inputs:
            if input_.value is None:
                raise ValueError("trying to construct command for task with missing input value")

            if input_.position is not None:
                positional_inputs.append(input_)
            else:
                nonpositional_inputs.append(input_)
        positional_inputs.sort(key=lambda i: i.position)

        if isinstance(self.base_command, list):
            command = self.base_command[:]
        else:
            command = [self.base_command]

        for input_ in positional_inputs:
            if input_.prefix is not None:
                command.append(input_.prefix)
            command.append(str(input_.value))
        for input_ in nonpositional_inputs:
            if input_.prefix is not None:
                command.append(input_.prefix)
            command.append(str(input_.value))

        # Append restart parameter and checkpoint file if CheckpointRequirement specified
        for hint in self.hints:
            if hint.class_ == "beeflow:CheckpointRequirement":
                if "bee_checkpoint_file__" in hint.params:
                    if "restart_parameters" in hint.params:
                        command.append(hint.params["restart_parameters"])
                    # Ignoring "container_path" for now
                    checkpoint_file = hint.params["bee_checkpoint_file__"]
                    # Charliecloud default bind mounts (this should taken from
                    # another requirement)
                    bind_mounts = {
                        os.getenv('HOME'): os.path.join('/home', os.getenv('USER')),
                    }
                    command.append(convert_path(checkpoint_file, bind_mounts))
                break

        return command



"""BeeApi wrapper around Flask-Restful for handling exceptions."""
import traceback
from flask import make_response, jsonify
from flask_restful import Api


class BeeApi(Api):
    """Wrapper around Flask-Restful's API to catch exceptions."""

    def handle_error(self, e):  # noqa (conflict on naming in base class vs. following convention)
        """Handle an error or exception."""
        return make_response(jsonify(error=traceback.format_exc()), 500)



"""This module contains the workflow action endpoints."""

from flask import make_response, jsonify
from flask_restful import Resource, reqparse
from beeflow.wf_manager.common import wf_db
from beeflow.common import log as bee_logging
from beeflow.wf_manager.resources import wf_utils


log = bee_logging.setup(__name__)


class WFActions(Resource):
    """Class to perform actions on existing workflows."""

    def __init__(self):
        """Initialize with passed json object."""
        self.reqparse = reqparse.RequestParser()

    def post(self, wf_id):
        """Start workflow. Send ready tasks to the task manager."""
        wfi = wf_utils.get_workflow_interface(wf_id)
        state = wfi.get_workflow_state()
        if state in ('RUNNING', 'PAUSED', 'COMPLETED'):
            resp = make_response(jsonify(msg='Cannot start workflow it is '
                                 f'{state.lower()}.',
                                         status='ok'), 200)
            return resp
        wfi.execute_workflow()
        tasks = wfi.get_ready_tasks()
        wf_utils.schedule_submit_tasks(wf_id, tasks)
        wf_id = wfi.workflow_id
        wf_utils.update_wf_status(wf_id, 'Running')
        wf_db.update_workflow_state(wf_id, 'Running')
        resp = make_response(jsonify(msg='Started workflow!', status='ok'), 200)
        return resp

    @staticmethod
    def get(wf_id):
        """Check the database for the current status of all tasks."""
        tasks = wf_db.get_tasks(wf_id)
        tasks_status = []
        if not tasks:
            log.info(f"Bad query for wf {wf_id}.")
            wf_status = 'No workflow with that ID is currently loaded'
            tasks_status.append('Unavailable')
            resp = make_response(jsonify(tasks_status=tasks_status,
                                 wf_status=wf_status, status='not found'), 404)

        for task in tasks:
            tasks_status.append(f"{task.name}--{task.status}")
        tasks_status = '\n'.join(tasks_status)
        wf_status = wf_utils.read_wf_status(wf_id)

        resp = make_response(jsonify(tasks_status=tasks_status,
                             wf_status=wf_status, status='ok'), 200)
        return resp

    @staticmethod
    def delete(wf_id):
        """Cancel the workflow. Lets current tasks finish running."""
        wfi = wf_utils.get_workflow_interface(wf_id)
        # Remove all tasks currently in the database
        if wfi.workflow_loaded():
            wfi.finalize_workflow()
        wf_utils.update_wf_status(wf_id, 'Cancelled')
        wf_db.update_workflow_state(wf_id, 'Cancelled')
        wf_db.delete_workflow(wf_id)
        log.info("Workflow cancelled")
        resp = make_response(jsonify(status='Cancelled'), 202)
        return resp

    def patch(self, wf_id):
        """Pause or resume workflow."""
        self.reqparse.add_argument('option', type=str, location='json')
        option = self.reqparse.parse_args()['option']

        wfi = wf_utils.get_workflow_interface(wf_id)
        wf_state = wfi.get_workflow_state()
        if option == 'pause' and wf_state == 'RUNNING':
            wfi.pause_workflow()
            wf_utils.update_wf_status(wf_id, 'Paused')
            wf_db.update_workflow_state(wf_id, 'Paused')
            log.info("Workflow Paused")
            resp = make_response(jsonify(status='Workflow Paused'), 200)
        elif option == 'resume' and wf_state == 'PAUSED':
            wfi.resume_workflow()
            tasks = wfi.get_ready_tasks()
            wf_utils.schedule_submit_tasks(wf_id, tasks)
            wf_utils.update_wf_status(wf_id, 'Running')
            wf_db.update_workflow_state(wf_id, 'Running')
            log.info("Workflow Paused")
            log.info("Workflow Resumed")
            resp = make_response(jsonify(status='Workflow Resumed'), 200)
        else:
            resp_msg = f'Cannot {option} workflow. It is currently {wf_state.lower()}.'
            log.info(resp_msg)
            resp = make_response(jsonify(status=resp_msg), 200)
        return resp
